# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import logging
import os
import time
import warnings
from typing import Union

import cudf
import dask_cudf
import numpy as np

from nemo_curator._compat import MINHASH_DEPRECATED_API, MINHASH_PERMUTED_AVAILABLE
from nemo_curator.datasets import DocumentDataset
from nemo_curator.log import create_logger
from nemo_curator.utils.distributed_utils import performance_report_if_with_ts_suffix


class MinHash:
    """
    Computes minhash signatures of a document corpus
    """

    def __init__(
        self,
        seed: int = 42,
        num_hashes: int = 260,
        char_ngrams: int = 24,
        use_64bit_hash: bool = False,
        logger: Union[logging.LoggerAdapter, str] = "./",
        id_field: str = "id",
        text_field: str = "text",
        profile_dir: str = None,
        cache_dir: str = None,
    ):
        """
        Parameters
        ----------
        seed: Seed for minhash permutations
        num_hashes: Length of minhash signature (No. of minhash permutations)
        char_ngrams: Width of text window (in characters) while computing minhashes.
        use_64bit_hash: Whether to use a 64 bit hash function.
        logger: Existing logger to log to, or a path to a log directory.
        id_field: Column in the Dataset denoting document ID.
        text_field: Column in the Dataset denoting document content.
        profile_dir: str, Default None
          If specified directory to write dask profile
        cache_dir: str, Default None
          If specified, will compute & write id, minhash pairs to directory
        """
        self.num_hashes = num_hashes
        self.char_ngram = char_ngrams

        if MINHASH_DEPRECATED_API:
            self.seeds = self.generate_seeds(n_seeds=self.num_hashes, seed=seed)
        else:
            self.seeds = self.generate_hash_permutation_seeds(
                bit_width=64 if use_64bit_hash else 32,
                n_permutations=self.num_hashes,
                seed=seed,
            )

        self.minhash_method = self.minhash64 if use_64bit_hash else self.minhash32
        self.id_field = id_field
        self.text_field = text_field

        if cache_dir is None and profile_dir is not None:
            warnings.warn(
                "cache_dir for intermediate outputs is required to generate profiles"
            )
        self.cache_dir = cache_dir
        self.profile_dir = profile_dir

        if isinstance(logger, str):
            self._logger = create_logger(
                rank=0,
                log_file=os.path.join(logger, "Minhash.log"),
                name="Minhash",
            )
        else:
            self._logger = logger

    def generate_seeds(self, n_seeds: int = 260, seed: int = 0) -> np.ndarray:
        """
        Generate seeds for all minhash permutations based on the given seed.
        """
        gen = np.random.RandomState(seed)
        return gen.randint(0, 1e6, size=n_seeds)

    def generate_hash_permutation_seeds(
        self, bit_width: int, n_permutations: int = 260, seed: int = 0
    ) -> np.ndarray:
        """
        Generate seeds for all minhash permutations based on the given seed.
        """
        gen = np.random.RandomState(seed)

        if bit_width == 32:
            MERSENNE_PRIME = np.uint32((1 << 31) - 1)
            dtype = np.uint32
        elif bit_width == 64:
            # For 64-bit, use a larger prime number suitable for 64-bit operations
            MERSENNE_PRIME = np.uint64((1 << 61) - 1)
            dtype = np.uint64
        else:
            raise ValueError("Unsupported bit width. Use either 32 or 64.")

        return np.array(
            [
                (
                    gen.randint(1, MERSENNE_PRIME, dtype=dtype),
                    gen.randint(0, MERSENNE_PRIME, dtype=dtype),
                )
                for _ in range(n_permutations)
            ],
            dtype=dtype,
        )

    def minhash32(
        self, ser: cudf.Series, seeds: np.ndarray, char_ngram: int
    ) -> cudf.Series:
        """
        Compute 32bit minhashes based on the MurmurHash3 algorithm
        """
        if not isinstance(ser, cudf.Series):
            raise TypeError("Expected data of type cudf.Series")

        if MINHASH_DEPRECATED_API:
            warnings.warn(
                "Using an outdated minhash implementation, please update to cuDF version 24.12 "
                "or later for improved performance. "
                "Install the latest version of cuDF using `pip install curator[cuda12x_nightly]`",
                category=FutureWarning,
            )
            seeds = cudf.Series(seeds, dtype="uint32")
            return ser.str.minhash(seeds=seeds, width=char_ngram)
        else:
            seeds_a = cudf.Series(seeds[:, 0], dtype="uint32")
            seeds_b = cudf.Series(seeds[:, 1], dtype="uint32")

            if MINHASH_PERMUTED_AVAILABLE:
                return ser.str.minhash_permuted(
                    a=seeds_a, b=seeds_b, seed=seeds[0][0], width=char_ngram
                )
            else:
                return ser.str.minhash(
                    a=seeds_a, b=seeds_b, seed=seeds[0][0], width=char_ngram
                )

    def minhash64(
        self, ser: cudf.Series, seeds: np.ndarray, char_ngram: int
    ) -> cudf.Series:
        """
        Compute 64bit minhashes based on the MurmurHash3 algorithm
        """
        if not isinstance(ser, cudf.Series):
            raise TypeError("Expected data of type cudf.Series")
        if MINHASH_DEPRECATED_API:
            warnings.warn(
                "Using an outdated minhash implementation, please update to cuDF version 24.12 "
                "or later for improved performance. "
                "Install the latest version of cuDF using `pip install curator[cuda12x_nightly]`",
                category=FutureWarning,
            )
            seeds = cudf.Series(seeds, dtype="uint64")
            return ser.str.minhash64(seeds=seeds, width=char_ngram)
        else:
            seeds_a = cudf.Series(seeds[:, 0], dtype="uint64")
            seeds_b = cudf.Series(seeds[:, 1], dtype="uint64")

            if MINHASH_PERMUTED_AVAILABLE:
                return ser.str.minhash64_permuted(
                    a=seeds_a, b=seeds_b, seed=seeds[0][0], width=char_ngram
                )
            else:
                return ser.str.minhash64(
                    a=seeds_a, b=seeds_b, seed=seeds[0][0], width=char_ngram
                )

    def __call__(self, dataset: DocumentDataset) -> Union[str, DocumentDataset]:
        """
        Computes the MinHash Signatures for a given dataset.
        Parameters
        ----------
        dataset: DocumentDataset
        The input datset to compute MinHashes.
        Returns
        -------
        DocumentDataset containing IDs of all documents and the corresponding MinHash Signature
        """
        result = dataset.df[[self.id_field]]
        result["_minhash_signature"] = dataset.df[self.text_field].map_partitions(
            self.minhash_method,
            seeds=self.seeds,
            char_ngram=self.char_ngram,
        )

        if self.cache_dir is None:
            return DocumentDataset(result)

        t0 = time.time()
        self._logger.info("Starting execution for Minhashes")
        write_path = os.path.join(self.cache_dir, "_minhashes.parquet")
        if os.path.exists(write_path):
            warnings.warn(
                f"Output path {write_path} already exists and will be overwritten"
            )
        with performance_report_if_with_ts_suffix(self.profile_dir, "minhash-profile"):
            result.to_parquet(write_path, write_index=False, overwrite=True)
        self._logger.info(
            f"Time taken for Minhash signature computation = {time.time() - t0}s and output written at {write_path}"
        )
        return DocumentDataset(
            dask_cudf.read_parquet(write_path, blocksize="2GB", aggregate_files=True)
        )



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .downstream_task import DownstreamTask, import_task
from .metrics import (
    ANLI,
    CB,
    COQA,
    MMLU,
    PIQA,
    RTE,
    WSC,
    ArcChallenge,
    ArcEasy,
    BigBenchHard,
    BigBenchLight,
    BoolQ,
    Copa,
    Drop,
    Lambada,
    Multilingual,
    MultiRC,
    NumDasc,
    OpenBookQA,
    Quac,
    Race,
    Record,
    Squad,
    StoryCloze,
    TriviaQA,
    WebQA,
    WiC,
    Winogrande,
)

__all__ = [
    "DownstreamTask",
    "import_task",
    "Race",
    "Squad",
    "ArcEasy",
    "ArcChallenge",
    "OpenBookQA",
    "BoolQ",
    "Copa",
    "RTE",
    "MultiRC",
    "WSC",
    "CB",
    "ANLI",
    "Record",
    "COQA",
    "TriviaQA",
    "Quac",
    "WebQA",
    "Drop",
    "WiC",
    "MMLU",
    "BigBenchHard",
    "BigBenchLight",
    "Multilingual",
    "PIQA",
    "Winogrande",
    "Lambada",
    "NumDasc",
    "StoryCloze",
]



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import asyncio
import logging
import os
from typing import Any, Coroutine, List, Optional, Tuple, Union

import yaml
from tqdm.asyncio import tqdm

from nemo_curator.log import create_logger
from nemo_curator.services.model_client import AsyncLLMClient
from nemo_curator.synthetic.error import YamlConversionError
from nemo_curator.synthetic.prompts import (
    DEFAULT_CLOSED_QA_PROMPT_TEMPLATE,
    DEFAULT_MACRO_TOPICS_PROMPT_TEMPLATE,
    DEFAULT_MATH_CLASSIFICATION_PROMPT_TEMPLATE,
    DEFAULT_MATH_MACRO_TOPICS_PROMPT_TEMPLATE,
    DEFAULT_MATH_SUBTOPICS_PROMPT_TEMPLATE,
    DEFAULT_OPEN_QA_FROM_TOPICS_PROMPT_TEMPLATE,
    DEFAULT_PYTHON_CLASSIFICATION_PROMPT_TEMPLATE,
    DEFAULT_PYTHON_MACRO_TOPICS_PROMPT_TEMPLATE,
    DEFAULT_PYTHON_SUBTOPICS_PROMPT_TEMPLATE,
    DEFAULT_REVISE_OPEN_QA_PROMPT_TEMPLATE,
    DEFAULT_REVISE_WRITING_TASK_PROMPT_TEMPLATE,
    DEFAULT_SUBTOPICS_PROMPT_TEMPLATE,
    DEFAULT_WRITING_TASK_PROMPT_TEMPLATE,
    DEFAULT_YAML_CONVERSION_PROMPT_TEMPLATE,
    DIALOGUE_NORMAL_USER_TURN_PROMPT_TEMPLATE,
    MATH_PROBLEM_GENERAL_PROMPT_TEMPLATE,
    PYTHON_PROBLEM_BEGINNER_PROMPT_TEMPLATE,
)


class AsyncNemotronGenerator:
    """
    Provides a collection of methods for generating synthetic data
    described in the Nemotron-4 340B Technical Report
    (https://arxiv.org/abs/2406.11704v1) and inspired by the
    UltraChat paper (https://arxiv.org/abs/2305.14233)
    """

    def __init__(
        self,
        llm_client: AsyncLLMClient,
        logger: Union[logging.LoggerAdapter, str] = "./",
        max_concurrent_requests: Optional[int] = None,
    ) -> None:
        self.client = llm_client
        self.max_concurrent_requests = max_concurrent_requests
        if isinstance(logger, str):
            self.logger = create_logger(
                rank=0,
                log_file=os.path.join(logger, "nemotron-generator.log"),
                name="AsyncNemotronGenrator",
            )
        else:
            self.logger = logger

    async def _prompt(
        self, model: str, prompt_template: str, prompt_kwargs: dict, model_kwargs: dict
    ) -> List[str]:
        prompt = prompt_template.format(**prompt_kwargs)
        messages = [{"role": "user", "content": prompt}]

        return await self.client.query_model(
            messages=messages, model=model, **model_kwargs
        )

    async def convert_response_to_yaml_list(
        self,
        llm_response: str,
        model: str,
        prompt_template: str = DEFAULT_YAML_CONVERSION_PROMPT_TEMPLATE,
        prompt_kwargs: dict = {},
        model_kwargs: dict = {},
    ) -> List[str]:
        """
        Converts a response of an LLM to a list of strings by querying an LLM
        Args:
            llm_response: The original unformatted response of the LLM
            model: The name of the model that should be used to generate the response.
                Must be available in the LLMClient passed in the constructor.
            prompt_template: A format string of the prompt to use. It must have a {llm_response}
                parameter that will be populated with the llm_response value passed in this function.
            prompt_kwargs: Any additional keyword arguments that should be passed to the prompt template.
                None are needed for the default template.
            model_kwargs: Any additional keyword arguments that should be passed to the LLMClient.query_model call.
        Returns:
            A parsed list of elements from the original LLM response
        """
        prompt_kwargs["llm_response"] = llm_response
        yaml_response = await self._prompt(
            model=model,
            prompt_template=prompt_template,
            prompt_kwargs=prompt_kwargs,
            model_kwargs=model_kwargs,
        )

        try:
            parsed_response = yaml.safe_load(yaml_response[0])
        except yaml.error.YAMLError as _:
            raise YamlConversionError(
                f"Error parsing yaml response: {yaml_response[0]}"
            )

        if not isinstance(parsed_response, list):
            raise YamlConversionError(
                f"Error: Parsed response was not a list: {parsed_response}"
            )

        for elem in parsed_response:
            if not isinstance(elem, str):
                raise YamlConversionError(
                    f"Error: Parsed response contains non-string elements in list: {parsed_response}"
                )
            if elem not in llm_response:
                raise YamlConversionError(
                    f"Conversion introduced hallucinations. Original response:\n{llm_response}\nConverted response:\n{parsed_response}\nHallucination:\n{elem}"
                )

        return parsed_response

    async def _try_convert_yaml_list(
        self,
        response: str,
        model: str,
        yaml_conversion_prompt_template: str,
        conversion_model_kwargs: dict,
        expected_length: int,
        ignore_conversion_failure: bool,
    ):
        try:
            parsed_list = await self.convert_response_to_yaml_list(
                response,
                model=model,
                prompt_template=yaml_conversion_prompt_template,
                model_kwargs=conversion_model_kwargs,
            )
            if len(parsed_list) != expected_length:
                raise YamlConversionError(
                    f"Error: Length of parsed list {len(parsed_list)} does not match expected length {expected_length}: {parsed_list}"
                )
        except YamlConversionError as e:
            if ignore_conversion_failure:
                return []
            else:
                raise e

        return parsed_list

    async def _gather(
        self, requests: List[Coroutine[Any, Any, List[str]]]
    ) -> List[str]:
        max_requests = self.max_concurrent_requests
        if max_requests is None:
            max_requests = len(requests)

        final_list = []
        for i in tqdm(range(0, len(requests), max_requests)):
            request_slice = requests[i : i + max_requests]
            result = await tqdm.gather(*request_slice)
            final_list.extend(result)

        return final_list

    async def generate_macro_topics(
        self,
        n_macro_topics: Union[int, str],
        model: str,
        prompt_template: str = DEFAULT_MACRO_TOPICS_PROMPT_TEMPLATE,
        prompt_kwargs: dict = {},
        model_kwargs: dict = {},
    ) -> List[str]:
        """
        Prompts an LLM to generate a list of macro topics about the world
        Args:
            n_macro_topics: The number of macro topics to generate.
            model: The name of the model that should be used to generate the macro topics.
                Must be available in the LLMClient passed in the constructor.
            prompt_template: A format string of the prompt to use. It must have the following parameters:
                - n_macro_topics: Will be populated with the n_macro_topics passed in this function
            prompt_kwargs: Any additional keyword arguments that should be passed to the prompt template.
                None are needed for the default template.
            model_kwargs: Any additional keyword arguments that should be passed to the LLMClient.query_model call.
        Returns:
            A list of responses from the LLM. The list is only greater than length 1 if n > 1 is set in model_kwargs.
        """
        prompt_kwargs["n_macro_topics"] = n_macro_topics
        macro_topics = await self._prompt(
            model=model,
            prompt_template=prompt_template,
            prompt_kwargs=prompt_kwargs,
            model_kwargs=model_kwargs,
        )

        return macro_topics

    async def generate_subtopics(
        self,
        macro_topic: str,
        n_subtopics: Union[int, str],
        model: str,
        prompt_template: str = DEFAULT_SUBTOPICS_PROMPT_TEMPLATE,
        prompt_kwargs: dict = {},
        model_kwargs: dict = {},
    ) -> List[str]:
        """
        Prompts an LLM to generate a list of subtopics relating to a macro topic
        Args:
            macro_topic: The macro topic to generate subtopics for.
            n_subtopics: The number of subtopics to generate per macro topic
            model: The name of the model that should be used to generate the response.
                Must be available in the LLMClient passed in the constructor.
            prompt_template: A format string of the prompt to use. It must have the following parameters:
                - n_subtopics: Will be populated with the n_subtopics passed in this function
                - macro_topic: Will be populated with the macro_topic passed in this function
            prompt_kwargs: Any additional keyword arguments that should be passed to the prompt template.
                None are needed for the default template.
            model_kwargs: Any additional keyword arguments that should be passed to the LLMClient.query_model call.
        Returns:
            A list of responses from the LLM. The list is only greater than length 1 if n > 1 is set in model_kwargs.
        """
        prompt_kwargs["n_subtopics"] = n_subtopics
        prompt_kwargs["macro_topic"] = macro_topic
        subtopics_response = await self._prompt(
            model=model,
            prompt_template=prompt_template,
            prompt_kwargs=prompt_kwargs,
            model_kwargs=model_kwargs,
        )

        return subtopics_response

    async def generate_open_qa_from_topic(
        self,
        topic: str,
        n_openlines: Union[str, int],
        model: str,
        prompt_template: str = DEFAULT_OPEN_QA_FROM_TOPICS_PROMPT_TEMPLATE,
        prompt_kwargs: dict = {},
        model_kwargs: dict = {},
    ) -> List[str]:
        """
        Prompts an LLM to generate a list of open Q&A questions based on a topic
        Args:
            topic: The topic to generate questions for.
            n_openlines: The number of questions to generate per topic.
            model: The name of the model that should be used to generate the response.
                Must be available in the LLMClient passed in the constructor.
            prompt_template: A format string of the prompt to use. It must have the following parameters:
                - n_openlines: Will be populated with the n_subtopics passed in this function
                - topic: Will be populated with the topic passed in this function
            prompt_kwargs: Any additional keyword arguments that should be passed to the prompt template.
                None are needed for the default template.
            model_kwargs: Any additional keyword arguments that should be passed to the LLMClient.query_model call.
        Returns:
            A list of responses from the LLM. The list is only greater than length 1 if n > 1 is set in model_kwargs.
        """
        prompt_kwargs["topic"] = topic
        prompt_kwargs["n_openlines"] = n_openlines
        openline_response = await self._prompt(
            model=model,
            prompt_template=prompt_template,
            prompt_kwargs=prompt_kwargs,
            model_kwargs=model_kwargs,
        )

        return openline_response

    async def revise_open_qa(
        self,
        openline: str,
        n_revisions: Union[str, int],
        model: str,
        prompt_template: str = DEFAULT_REVISE_OPEN_QA_PROMPT_TEMPLATE,
        prompt_kwargs: dict = {},
        model_kwargs: dict = {},
    ) -> List[str]:
        """
        Prompts an LLM to revise an open Q&A question a given number of times
        Args:
            openline: An openline to revise
            n_revisions: The number of revisions to generate for the question.
            model: The name of the model that should be used to generate the response.
                Must be available in the LLMClient passed in the constructor.
            prompt_template: A format string of the prompt to use. It must have the following parameters:
                - openline: Will be populated with the openline passed in this function
                - n_revisions: Will be populated with the n_revisions passed in this function
            prompt_kwargs: Any additional keyword arguments that should be passed to the prompt template.
                None are needed for the default template.
            model_kwargs: Any additional keyword arguments that should be passed to the LLMClient.query_model call.
        Returns:
            A list of responses from the LLM. The list is only greater than length 1 if n > 1 is set in model_kwargs.
        """
        prompt_kwargs["openline"] = openline
        prompt_kwargs["n_revisions"] = n_revisions
        revisions = await self._prompt(
            model=model,
            prompt_template=prompt_template,
            prompt_kwargs=prompt_kwargs,
            model_kwargs=model_kwargs,
        )

        return revisions

    async def generate_writing_tasks(
        self,
        topic: str,
        text_material_type: str,
        n_openlines: Union[str, int],
        model: str,
        prompt_template: str = DEFAULT_WRITING_TASK_PROMPT_TEMPLATE,
        prompt_kwargs: dict = {},
        model_kwargs: dict = {},
    ) -> List[str]:
        """
        Prompts an LLM to generate a list of writing tasks based on a topic and document type
        Args:
            topic: The topic to generate writing tasks for.
            text_material_type: The type of the document the question should ask to generate (e.g., "Email", "Poem")
            n_openlines: The number of tasks to generate per topic and text material pair.
            model: The name of the model that should be used to generate the response.
                Must be available in the LLMClient passed in the constructor.
            prompt_template: A format string of the prompt to use. It must have the following parameters:
                - topic: Will be populated with the topic passed in this function
                - text_material_type: Will be populated with the text_material_type passed in this function
                - n_openlines: Will be populated with the n_openlines passed in this function
            prompt_kwargs: Any additional keyword arguments that should be passed to the prompt template.
                None are needed for the default template.
            model_kwargs: Any additional keyword arguments that should be passed to the LLMClient.query_model call.
        Returns:
            A list of responses from the LLM. The list is only greater than length 1 if n > 1 is set in model_kwargs.
        """
        prompt_kwargs["topic"] = topic
        prompt_kwargs["text_material_type"] = text_material_type
        prompt_kwargs["n_openlines"] = n_openlines
        writing_tasks = await self._prompt(
            model=model,
            prompt_template=prompt_template,
            prompt_kwargs=prompt_kwargs,
            model_kwargs=model_kwargs,
        )

        return writing_tasks

    async def revise_writing_tasks(
        self,
        openline: str,
        n_revisions: Union[str, int],
        model: str,
        prompt_template: str = DEFAULT_REVISE_WRITING_TASK_PROMPT_TEMPLATE,
        prompt_kwargs: dict = {},
        model_kwargs: dict = {},
    ) -> List[str]:
        """
        Prompts an LLM to revise a writing task a given number of times
        Args:
            openline: An openline to revise
            n_revisions: The number of revisions to generate for the task.
            model: The name of the model that should be used to generate the response.
                Must be available in the LLMClient passed in the constructor.
            prompt_template: A format string of the prompt to use. It must have the following parameters:
                - openline: Will be populated with the openline passed in this function
                - n_revisions: Will be populated with the n_revisions passed in this function
            prompt_kwargs: Any additional keyword arguments that should be passed to the prompt template.
                None are needed for the default template.
            model_kwargs: Any additional keyword arguments that should be passed to the LLMClient.query_model call.
        Returns:
            A list of responses from the LLM. The list is only greater than length 1 if n > 1 is set in model_kwargs.
        """
        prompt_kwargs["openline"] = openline
        prompt_kwargs["n_revisions"] = n_revisions
        revisions = await self._prompt(
            model=model,
            prompt_template=prompt_template,
            prompt_kwargs=prompt_kwargs,
            model_kwargs=model_kwargs,
        )

        return revisions

    async def generate_closed_qa_instructions(
        self,
        document: str,
        n_openlines: Union[str, int],
        model: str,
        prompt_template: str = DEFAULT_CLOSED_QA_PROMPT_TEMPLATE,
        prompt_kwargs: dict = {},
        model_kwargs: dict = {},
    ) -> List[str]:
        """
        Prompts an LLM to generate a list of closed Q&A questions based on a reference document
        Args:
            document: The document to use when generating questions
            n_openlines: The number of questions to generate per document.
            model: The name of the model that should be used to generate the response.
                Must be available in the LLMClient passed in the constructor.
            prompt_template: A format string of the prompt to use. It must have the following parameters:
                - document: Will be populated with the document passed in this function
                - n_openlines: Will be populated with the n_openlines passed in this function
            prompt_kwargs: Any additional keyword arguments that should be passed to the prompt template.
                None are needed for the default template.
            model_kwargs: Any additional keyword arguments that should be passed to the LLMClient.query_model call.
        Returns:
            A list of responses from the LLM. The list is only greater than length 1 if n > 1 is set in model_kwargs.
        """
        prompt_kwargs["document"] = document
        prompt_kwargs["n_openlines"] = n_openlines
        openline_response = await self._prompt(
            model=model,
            prompt_template=prompt_template,
            prompt_kwargs=prompt_kwargs,
            model_kwargs=model_kwargs,
        )

        return openline_response

    async def generate_math_macro_topics(
        self,
        n_macro_topics: Union[int, str],
        school_level: str,
        model: str,
        prompt_template: str = DEFAULT_MATH_MACRO_TOPICS_PROMPT_TEMPLATE,
        prompt_kwargs: dict = {},
        model_kwargs: dict = {},
    ) -> List[str]:
        """
        Prompts an LLM to generate a list of macro topics about math
        Args:
            n_macro_topics: The number of macro topics to generate. Can be an integer like 5 or a string like "five".
            school_level: The school level the math questions should be targeted at.
            model: The name of the model that should be used to generate the macro topics.
                Must be available in the LLMClient passed in the constructor.
            prompt_template: A format string of the prompt to use. It must have the following parameters:
                - n_macro_topics: Will be populated with the n_macro_topics passed in this function
                - school_level: Will be populated with the school_level passed in this function
            prompt_kwargs: Any additional keyword arguments that should be passed to the prompt template.
                None are needed for the default template.
            model_kwargs: Any additional keyword arguments that should be passed to the LLMClient.query_model call.
        Returns:
            A list of responses from the LLM. The list is only greater than length 1 if n > 1 is set in model_kwargs.
        """
        prompt_kwargs["n_macro_topics"] = n_macro_topics
        prompt_kwargs["school_level"] = school_level
        macro_topics = await self._prompt(
            model=model,
            prompt_template=prompt_template,
            prompt_kwargs=prompt_kwargs,
            model_kwargs=model_kwargs,
        )

        return macro_topics

    async def generate_math_subtopics(
        self,
        macro_topic: str,
        n_subtopics: Union[int, str],
        model: str,
        prompt_template: str = DEFAULT_MATH_SUBTOPICS_PROMPT_TEMPLATE,
        prompt_kwargs: dict = {},
        model_kwargs: dict = {},
    ) -> List[str]:
        """
        Prompts an LLM to generate a list of subtopics relating to a math macro topic
        Args:
            macro_topic: The macro topic to generate subtopics for.
            n_subtopics: The number of subtopics to generate per macro topic
            model: The name of the model that should be used to generate the response.
                Must be available in the LLMClient passed in the constructor.
            prompt_template: A format string of the prompt to use. It must have the following parameters:
                - n_subtopics: Will be populated with the n_subtopics passed in this function
                - macro_topic: Will be populated with the macro_topic passed in this function
            prompt_kwargs: Any additional keyword arguments that should be passed to the prompt template.
                None are needed for the default template.
            model_kwargs: Any additional keyword arguments that should be passed to the LLMClient.query_model call.
        Returns:
            A list of responses from the LLM. The list is only greater than length 1 if n > 1 is set in model_kwargs.
        """
        prompt_kwargs["n_subtopics"] = n_subtopics
        prompt_kwargs["macro_topic"] = macro_topic
        subtopics_response = await self._prompt(
            model=model,
            prompt_template=prompt_template,
            prompt_kwargs=prompt_kwargs,
            model_kwargs=model_kwargs,
        )

        return subtopics_response

    async def classify_math_entity(
        self,
        entity: str,
        model: str,
        prompt_template: str = DEFAULT_MATH_CLASSIFICATION_PROMPT_TEMPLATE,
        prompt_kwargs: dict = {},
        model_kwargs={},
    ) -> List[str]:
        """
        Prompts an LLM to classify if an entity is related to math
        Args:
            entity: The entity to classify
            model: The name of the model that should be used to generate the response.
                Must be available in the LLMClient passed in the constructor.
            prompt_template: A format string of the prompt to use. It must have the following parameters:
                - entity: Will be populated with the entity passed in this function
            prompt_kwargs: Any additional keyword arguments that should be passed to the prompt template.
                None are needed for the default template.
            model_kwargs: Any additional keyword arguments that should be passed to the LLMClient.query_model call.
        Returns:
            A list of responses from the LLM. The list is only greater than length 1 if n > 1 is set in model_kwargs.
        """
        prompt_kwargs["entity"] = entity
        classification_response = await self._prompt(
            model=model,
            prompt_template=prompt_template,
            prompt_kwargs=prompt_kwargs,
            model_kwargs=model_kwargs,
        )

        return classification_response

    async def generate_math_problem(
        self,
        topic: str,
        n_openlines: Union[str, int],
        model: str,
        prompt_template: str = MATH_PROBLEM_GENERAL_PROMPT_TEMPLATE,
        prompt_kwargs: dict = {},
        model_kwargs: dict = {},
    ) -> List[str]:
        """
        Prompts an LLM to generate a list of math problems based on a topic
        Args:
            topic: The topic to generate problems for.
            n_openlines: The number of problems to generate per topic.
            model: The name of the model that should be used to generate the response.
                Must be available in the LLMClient passed in the constructor.
            prompt_template: A format string of the prompt to use. It must have the following parameters:
                - n_openlines: Will be populated with the n_subtopics passed in this function
                - topic: Will be populated with the topic passed in this function
                Some example templates found in nemo_curator.synthetic include:
                - MATH_PROBLEM_GENERAL_PROMPT_TEMPLATE
                - MATH_PROBLEM_BEGINNER_PROMPT_TEMPLATE
            prompt_kwargs: Any additional keyword arguments that should be passed to the prompt template.
                None are needed for the default template.
            model_kwargs: Any additional keyword arguments that should be passed to the LLMClient.query_model call.
        Returns:
            A list of responses from the LLM. The list is only greater than length 1 if n > 1 is set in model_kwargs.
        """
        prompt_kwargs["topic"] = topic
        prompt_kwargs["n_openlines"] = n_openlines
        openline_response = await self._prompt(
            model=model,
            prompt_template=prompt_template,
            prompt_kwargs=prompt_kwargs,
            model_kwargs=model_kwargs,
        )

        return openline_response

    async def generate_python_macro_topics(
        self,
        n_macro_topics: Union[int, str],
        model: str,
        prompt_template: str = DEFAULT_PYTHON_MACRO_TOPICS_PROMPT_TEMPLATE,
        prompt_kwargs: dict = {},
        model_kwargs: dict = {},
    ) -> List[str]:
        """
        Prompts an LLM to generate a list of macro topics about the Python programming language
        Args:
            n_macro_topics: The number of macro topics to generate. Can be an integer like 5 or a string like "five".
            model: The name of the model that should be used to generate the macro topics.
                Must be available in the LLMClient passed in the constructor.
            prompt_template: A format string of the prompt to use. It must have the following parameters:
                - n_macro_topics: Will be populated with the n_macro_topics passed in this function
            prompt_kwargs: Any additional keyword arguments that should be passed to the prompt template.
                None are needed for the default template.
            model_kwargs: Any additional keyword arguments that should be passed to the LLMClient.query_model call.
        Returns:
            A list of responses from the LLM. The list is only greater than length 1 if n > 1 is set in model_kwargs.
        """
        prompt_kwargs["n_macro_topics"] = n_macro_topics
        macro_topics = await self._prompt(
            model=model,
            prompt_template=prompt_template,
            prompt_kwargs=prompt_kwargs,
            model_kwargs=model_kwargs,
        )

        return macro_topics

    async def generate_python_subtopics(
        self,
        macro_topic: str,
        n_subtopics: Union[int, str],
        model: str,
        prompt_template: str = DEFAULT_PYTHON_SUBTOPICS_PROMPT_TEMPLATE,
        prompt_kwargs: dict = {},
        model_kwargs: dict = {},
    ) -> List[str]:
        """
        Prompts an LLM to generate a list of subtopics relating to a Python macro topic
        Args:
            macro_topic: The macro topic to generate subtopics for.
            n_subtopics: The number of subtopics to generate per macro topic
            model: The name of the model that should be used to generate the response.
                Must be available in the LLMClient passed in the constructor.
            prompt_template: A format string of the prompt to use. It must have the following parameters:
                - n_subtopics: Will be populated with the n_subtopics passed in this function
                - macro_topic: Will be populated with the macro_topic passed in this function
            prompt_kwargs: Any additional keyword arguments that should be passed to the prompt template.
                None are needed for the default template.
            model_kwargs: Any additional keyword arguments that should be passed to the LLMClient.query_model call.
        Returns:
            A list of responses from the LLM. The list is only greater than length 1 if n > 1 is set in model_kwargs.
        """
        prompt_kwargs["n_subtopics"] = n_subtopics
        prompt_kwargs["macro_topic"] = macro_topic
        subtopics_response = await self._prompt(
            model=model,
            prompt_template=prompt_template,
            prompt_kwargs=prompt_kwargs,
            model_kwargs=model_kwargs,
        )

        return subtopics_response

    async def classify_python_entity(
        self,
        entity: str,
        model: str,
        prompt_template: str = DEFAULT_PYTHON_CLASSIFICATION_PROMPT_TEMPLATE,
        prompt_kwargs: dict = {},
        model_kwargs: dict = {},
    ) -> List[str]:
        """
        Prompts an LLM to classify if an entity is related to Python
        Args:
            entity: The entity to classify
            model: The name of the model that should be used to generate the response.
                Must be available in the LLMClient passed in the constructor.
            prompt_template: A format string of the prompt to use. It must have the following parameters:
                - entity: Will be populated with the entity passed in this function
            prompt_kwargs: Any additional keyword arguments that should be passed to the prompt template.
                None are needed for the default template.
            model_kwargs: Any additional keyword arguments that should be passed to the LLMClient.query_model call.
        Returns:
            A list of responses from the LLM. The list is only greater than length 1 if n > 1 is set in model_kwargs.
        """
        prompt_kwargs["entity"] = entity
        classification_response = await self._prompt(
            model=model,
            prompt_template=prompt_template,
            prompt_kwargs=prompt_kwargs,
            model_kwargs=model_kwargs,
        )

        return classification_response

    async def generate_python_problem(
        self,
        topic: str,
        n_openlines: Union[str, int],
        model: str,
        language="Python",
        prompt_template: str = PYTHON_PROBLEM_BEGINNER_PROMPT_TEMPLATE,
        prompt_kwargs: dict = {},
        model_kwargs: dict = {},
    ) -> List[str]:
        """
        Prompts an LLM to generate a list of coding problems based on a topic
        Args:
            topic: The topic to generate problems for.
            n_openlines: The number of problems to generate per topic.
            model: The name of the model that should be used to generate the response.
                Must be available in the LLMClient passed in the constructor.
            language: The programming language to target when generating these questions.
            prompt_template: A format string of the prompt to use. It must have the following parameters:
                - n_openlines: Will be populated with the n_subtopics passed in this function
                - topic: Will be populated with the topic passed in this function
                - language: Will be populated with the language passed in this function
                Some example templates found in nemo_curator.synthetic include:
                - PYTHON_PROBLEM_BEGINNER_PROMPT_TEMPLATE
                - PYTHON_PROBLEM_INTERMEDIATE_PROMPT_TEMPLATE
                - PYTHON_PROBLEM_ADVANCED_PROMPT_TEMPLATE
            prompt_kwargs: Any additional keyword arguments that should be passed to the prompt template.
                None are needed for the default template.
            model_kwargs: Any additional keyword arguments that should be passed to the LLMClient.query_model call.
        Returns:
            A list of responses from the LLM. The list is only greater than length 1 if n > 1 is set in model_kwargs.
        """
        prompt_kwargs["topic"] = topic
        prompt_kwargs["n_openlines"] = n_openlines
        prompt_kwargs["language"] = language
        openline_response = await self._prompt(
            model=model,
            prompt_template=prompt_template,
            prompt_kwargs=prompt_kwargs,
            model_kwargs=model_kwargs,
        )

        return openline_response

    async def generate_dialogue(
        self,
        openline: str,
        user_model: str,
        assistant_model: str,
        n_user_turns: int = 3,
        prompt_template: str = DIALOGUE_NORMAL_USER_TURN_PROMPT_TEMPLATE,
        prompt_kwargs: dict = {},
        user_model_kwargs: dict = {},
        assistant_model_kwargs: dict = {},
    ) -> List[dict]:
        """
        Prompts an LLM to generate a dialogue based on a given openline.
        The LLM will alternate impersonating the user and the assistant.
        Args:
            openline: The openline that will comprise the first user turn.
            user_model: The model that will be impersonating the user.
                Must be available in the LLMClient passed in the constructor.
            assistant_model: The model that will be impersonating the assistant
                Must be available in the LLMClient passed in the constructor.
            n_user_turns: The number of user turns to go through. The openline counts as 1 user turn.
                Therefore, if there are 3 user turns, 2 will be generated by the LLM impersonating the user.
            prompt_template: A format string of the prompt to use when impersonating the user.
                It must have the following parameters:
                - converstation_history: Will be populated with a formatted history of the dialogue up to that point.
                Some example templates found in nemo_curator.synthetic include:
                - DIALOGUE_NORMAL_USER_TURN_PROMPT_TEMPLATE
                - DIALOGUE_COMPLEX_USER_TURN_PROMPT_TEMPLATE
                - DIALOGUE_CONCISE_USER_TURN_PROMPT_TEMPLATE
            prompt_kwargs: Any additional keyword arguments that should be passed to the prompt template.
                None are needed for the default template.
            user_model_kwargs: Any additional keyword arguments that should be passed to the
                LLMClient.query_model call for the user.
            assistant_model_kwargs: Any additional keyword arguments that should be passed to the
                LLMClient.query_model call for the assistant.
        Returns:
            A conversation between a User and Assistant
        """
        conversation_history = [{"role": "user", "content": openline}]
        first_assistant_response = await self.client.query_model(
            messages=conversation_history,
            model=assistant_model,
            **assistant_model_kwargs,
        )
        first_assistant_response = first_assistant_response[0]
        conversation_history.append(
            {"role": "assistant", "content": first_assistant_response}
        )
        for _ in range(n_user_turns - 1):
            user_response = await self._impersonate_user(
                conversation_history=conversation_history,
                model=user_model,
                prompt_template=prompt_template,
                prompt_kwargs=prompt_kwargs,
                model_kwargs=user_model_kwargs,
            )
            conversation_history.append({"role": "user", "content": user_response})
            assistant_response = await self.client.query_model(
                messages=conversation_history,
                model=assistant_model,
                **assistant_model_kwargs,
            )
            assistant_response = assistant_response[0]
            conversation_history.append(
                {"role": "assistant", "content": assistant_response}
            )

        return conversation_history

    async def generate_two_turn_prompt(
        self,
        openline: str,
        user_model: str,
        assistant_model: str,
        prompt_template: str = DIALOGUE_NORMAL_USER_TURN_PROMPT_TEMPLATE,
        prompt_kwargs: dict = {},
        user_model_kwargs: dict = {},
        assistant_model_kwargs: dict = {},
    ) -> List[dict]:
        """
        Prompts an LLM to generate a response as an assistant, then as the user based on a given openline.
        The conversation will look like "User -> Assistant -> User"
        Args:
            openline: The openline that will comprise the first user turn.
            user_model: The model that will be impersonating the user.
                Must be available in the LLMClient passed in the constructor.
            assistant_model: The model that will be impersonating the assistant
                Must be available in the LLMClient passed in the constructor.
            prompt_template: A format string of the prompt to use when impersonating the user.
                It must have the following parameters:
                - converstation_history: Will be populated with a formatted history of the dialogue up to that point.
                Some example templates found in nemo_curator.synthetic include:
                - DIALOGUE_NORMAL_USER_TURN_PROMPT_TEMPLATE
                - DIALOGUE_COMPLEX_USER_TURN_PROMPT_TEMPLATE
                - DIALOGUE_CONCISE_USER_TURN_PROMPT_TEMPLATE
            prompt_kwargs: Any additional keyword arguments that should be passed to the prompt template.
                None are needed for the default template.
            user_model_kwargs: Any additional keyword arguments that should be passed to the
                LLMClient.query_model call for the user.
            assistant_model_kwargs: Any additional keyword arguments that should be passed to the
                LLMClient.query_model call for the assistant.
        Returns:
            A conversation between a User and Assistant
        """
        conversation_history = [{"role": "user", "content": openline}]
        first_assistant_response = await self.client.query_model(
            messages=conversation_history,
            model=assistant_model,
            **assistant_model_kwargs,
        )
        first_assistant_response = first_assistant_response[0]
        conversation_history.append(
            {"role": "assistant", "content": first_assistant_response}
        )

        user_response = await self._impersonate_user(
            conversation_history=conversation_history,
            model=user_model,
            prompt_template=prompt_template,
            prompt_kwargs=prompt_kwargs,
            model_kwargs=user_model_kwargs,
        )
        conversation_history.append({"role": "user", "content": user_response})

        return conversation_history

    async def _impersonate_user(
        self,
        conversation_history: List[dict],
        model: str,
        prompt_template: str,
        prompt_kwargs: dict,
        model_kwargs: dict,
    ) -> str:
        # Convert the conversation history to a string
        history_str = ""
        for turn in conversation_history:
            history_str += f"{turn['role'].capitalize()}: {turn['content']}"
        prompt_kwargs["conversation_history"] = history_str
        response = await self._prompt(
            model=model,
            prompt_template=prompt_template,
            prompt_kwargs=prompt_kwargs,
            model_kwargs=model_kwargs,
        )

        return response[0]

    async def run_open_qa_pipeline(
        self,
        n_macro_topics: Union[str, int],
        n_subtopics: Union[str, int],
        n_openlines: Union[str, int],
        n_revisions: Union[str, int],
        model: str,
        macro_topic_prompt_template: str = DEFAULT_MACRO_TOPICS_PROMPT_TEMPLATE,
        subtopic_prompt_template: str = DEFAULT_SUBTOPICS_PROMPT_TEMPLATE,
        open_qa_from_topics_prompt_template: str = DEFAULT_OPEN_QA_FROM_TOPICS_PROMPT_TEMPLATE,
        revise_open_qa_prompt_template: str = DEFAULT_REVISE_OPEN_QA_PROMPT_TEMPLATE,
        yaml_conversion_prompt_template: str = DEFAULT_YAML_CONVERSION_PROMPT_TEMPLATE,
        base_model_kwargs: dict = {},
        conversion_model_kwargs: dict = {},
        additional_macro_topics: List[str] = [],
        additional_subtopics: List[str] = [],
        ignore_conversion_failure: bool = False,
        combine_topics: bool = True,
    ) -> List[str]:
        """
        Runs a pipeline for automatically generating Open Q&A openlines for a dialogue
        Args:
            n_macro_topics: The number of macro topics to generate
            n_subtopics: The number of subtopics to generate per macro topic
            n_openlines: The number of questions to generate per topic.
            n_revisions: The number of revisions to generate per original question.
            model: The name of the model that should be used to generate all the responses.
                Must be available in the LLMClient passed in the constructor.
            macro_topic_prompt_template: A format string of the prompt to use. It must have the following parameters:
                - n_macro_topics: Will be populated with the n_macro_topics passed in this function
                No additional parameters may be passed to this prompt template.
            subtopic_prompt_template: A format string of the prompt to use. It must have the following parameters:
                - n_subtopics: Will be populated with the n_subtopics passed in this function
                - macro_topic: Will be populated with a generated macro topic
                No additional parameters may be passed to this prompt template.
            open_qa_from_topics_prompt_template: A format string of the prompt to use. It must have the following parameters:
                - n_openlines: Will be populated with the n_openlines passed in this function
                - topic: Will be populated with a generated topic
                No additional parameters may be passed to this prompt template.
            revise_open_qa_prompt_template: A format string of the prompt to use. It must have the following parameters:
                - n_revisions: Will be populated with the n_revisions passed in this function
                - openline: Will be populated with a generated open Q&A openline
                No additional parameters may be passed to this prompt template.
            yaml_conversion_prompt_template: A format string of the prompt to use. It must have the following parameters:
                - llm_response: Will be populated with the raw LLM response from each stage of the pipeline
                No additional parameters may be passed to this prompt template.
            base_model_kwargs: Any additional keyword arguments that should be passed to the
                LLMClient.query_model call for the normal stages of the pipeline.
            conversion_model_kwargs: Any additional keyword arguments that should be passed to the
                LLMClient.query_model call for the yaml conversion stages of the pipeline.
            ignore_conversion_failure: Ignores yaml conversion failures when able and discards the data
                that conversion was attempted on
            combine_topics: If True, mixes the macro topics with the subtopics when generating openlines.
                If False, only the subtopics are used.
        Returns:
            A list of synthetically generated open Q&A prompts
        """
        self.logger.info("Starting open q&a pipeline")
        # Generate the macro topics
        self.logger.info("Starting macro topic generation")
        responses = await self.generate_macro_topics(
            n_macro_topics=n_macro_topics,
            model=model,
            model_kwargs=base_model_kwargs,
            prompt_template=macro_topic_prompt_template,
        )

        try:
            macro_topics = await self.convert_response_to_yaml_list(
                responses[0],
                model=model,
                prompt_template=yaml_conversion_prompt_template,
                model_kwargs=conversion_model_kwargs,
            )
            if len(macro_topics) != n_macro_topics and not ignore_conversion_failure:
                raise YamlConversionError(
                    f"Error: Length of macro topics {len(macro_topics)} does not match desired n_macro_topics {n_macro_topics}: {macro_topics}"
                )
        except YamlConversionError as e:
            if ignore_conversion_failure:
                macro_topics = []
            else:
                raise e

        macro_topics.extend(additional_macro_topics)
        self.logger.info("Finished macro topic generation")

        # Generate the subtopics
        raw_topics = [
            self._generate_parse_subtopic(
                macro_topic=macro_topic,
                n_subtopics=n_subtopics,
                model=model,
                subtopic_prompt_template=subtopic_prompt_template,
                yaml_conversion_prompt_template=yaml_conversion_prompt_template,
                base_model_kwargs=base_model_kwargs,
                conversion_model_kwargs=conversion_model_kwargs,
                ignore_conversion_failure=ignore_conversion_failure,
            )
            for macro_topic in macro_topics
        ]
        self.logger.info("Starting subtopic generation")
        raw_topics = await self._gather(raw_topics)
        topic_list = [item for subtopics in raw_topics for item in subtopics]
        topic_list.extend(additional_subtopics)
        self.logger.info("Finished subtopic generation")

        # Mix the macro topics with the subtopics
        if combine_topics:
            topic_list.extend(macro_topics)

        # Generate the openlines
        raw_lines = [
            self._generate_parse_openline(
                subtopic=subtopic,
                n_openlines=n_openlines,
                model=model,
                open_qa_from_topics_prompt_template=open_qa_from_topics_prompt_template,
                yaml_conversion_prompt_template=yaml_conversion_prompt_template,
                base_model_kwargs=base_model_kwargs,
                conversion_model_kwargs=conversion_model_kwargs,
                ignore_conversion_failure=ignore_conversion_failure,
            )
            for subtopic in topic_list
        ]
        self.logger.info("Starting openline generation")
        raw_lines = await self._gather(raw_lines)
        openlines = [item for lines in raw_lines for item in lines]
        self.logger.info("Finished openline generation")

        # Revise the openlines
        raw_revisions = [
            self._revise_parse_openline(
                openline=openline,
                n_revisions=n_revisions,
                model=model,
                revise_open_qa_prompt_template=revise_open_qa_prompt_template,
                yaml_conversion_prompt_template=yaml_conversion_prompt_template,
                base_model_kwargs=base_model_kwargs,
                conversion_model_kwargs=conversion_model_kwargs,
                ignore_conversion_failure=ignore_conversion_failure,
            )
            for openline in openlines
        ]
        self.logger.info("Starting openline revision")
        raw_revisions = await self._gather(raw_revisions)
        revised_openlines = [item for revisions in raw_revisions for item in revisions]
        self.logger.info("Finished openline revision")
        self.logger.info("Finished open q&a pipeline")

        return revised_openlines

    async def _generate_parse_subtopic(
        self,
        macro_topic: str,
        n_subtopics: Union[int, str],
        model: str,
        subtopic_prompt_template: str,
        yaml_conversion_prompt_template: str,
        base_model_kwargs: dict,
        conversion_model_kwargs: dict,
        ignore_conversion_failure: bool,
    ) -> List[str]:
        subtopic = await self.generate_subtopics(
            macro_topic=macro_topic,
            n_subtopics=n_subtopics,
            model=model,
            model_kwargs=base_model_kwargs,
            prompt_template=subtopic_prompt_template,
        )
        subtopic = subtopic[0]
        return await self._try_convert_yaml_list(
            subtopic,
            model=model,
            yaml_conversion_prompt_template=yaml_conversion_prompt_template,
            conversion_model_kwargs=conversion_model_kwargs,
            expected_length=n_subtopics,
            ignore_conversion_failure=ignore_conversion_failure,
        )

    async def _generate_parse_openline(
        self,
        subtopic: str,
        n_openlines: Union[int, str],
        model: str,
        open_qa_from_topics_prompt_template: str,
        yaml_conversion_prompt_template: str,
        base_model_kwargs: dict,
        conversion_model_kwargs: dict,
        ignore_conversion_failure: bool,
    ) -> List[str]:
        openline = await self.generate_open_qa_from_topic(
            topic=subtopic,
            n_openlines=n_openlines,
            model=model,
            model_kwargs=base_model_kwargs,
            prompt_template=open_qa_from_topics_prompt_template,
        )
        openline = openline[0]
        return await self._try_convert_yaml_list(
            openline,
            model=model,
            yaml_conversion_prompt_template=yaml_conversion_prompt_template,
            conversion_model_kwargs=conversion_model_kwargs,
            expected_length=n_openlines,
            ignore_conversion_failure=ignore_conversion_failure,
        )

    async def _revise_parse_openline(
        self,
        openline: str,
        n_revisions: Union[int, str],
        model: str,
        revise_open_qa_prompt_template: str,
        yaml_conversion_prompt_template: str,
        base_model_kwargs: dict,
        conversion_model_kwargs: dict,
        ignore_conversion_failure: bool,
    ) -> List[str]:
        revised_openline = await self.revise_open_qa(
            openline=openline,
            n_revisions=n_revisions,
            model=model,
            model_kwargs=base_model_kwargs,
            prompt_template=revise_open_qa_prompt_template,
        )
        revised_openline = revised_openline[0]
        return await self._try_convert_yaml_list(
            revised_openline,
            model=model,
            yaml_conversion_prompt_template=yaml_conversion_prompt_template,
            conversion_model_kwargs=conversion_model_kwargs,
            expected_length=n_revisions,
            ignore_conversion_failure=ignore_conversion_failure,
        )

    async def run_writing_pipeline(
        self,
        topics: List[str],
        text_material_types: List[str],
        n_openlines: Union[str, int],
        n_revisions: Union[str, int],
        model: str,
        writing_task_prompt_template: str = DEFAULT_WRITING_TASK_PROMPT_TEMPLATE,
        revise_writing_task_prompt_template: str = DEFAULT_REVISE_WRITING_TASK_PROMPT_TEMPLATE,
        yaml_conversion_prompt_template: str = DEFAULT_YAML_CONVERSION_PROMPT_TEMPLATE,
        base_model_kwargs: dict = {},
        conversion_model_kwargs: dict = {},
        ignore_conversion_failure: bool = False,
    ) -> List[str]:
        """
        Runs a pipeline for automatically generating writing task openlines for a dialogue
        Args:
            topics: A list of topics to generate tasks for
            text_material_types: A list of writing material types, like "Essay" or "Blog post"
            n_openlines: The number of tasks to generate per (topic, text_material_type) pair.
            n_revisions: The number of revisions to generate per original task.
            model: The name of the model that should be used to generate all the responses.
                Must be available in the LLMClient passed in the constructor.
            writing_task_prompt_template: A format string of the prompt to use. It must have the following parameters:
                - n_openlines: Will be populated with the n_openlines passed in this function
                - topic: Will be populated with one element of the topics list passed in this function
                - text_material_type: Will be populated with one element of the text_material_types list passed in this function
                No additional parameters may be passed to this prompt template.
            revise_writing_task_prompt_template: A format string of the prompt to use. It must have the following parameters:
                - n_revisions: Will be populated with the n_revisions passed in this function
                - openline: Will be populated with one of the writing tasks generated in the pipeline.
                No additional parameters may be passed to this prompt template.
            yaml_conversion_prompt_template: A format string of the prompt to use. It must have the following parameters:
                - llm_response: Will be populated with the raw LLM response from each stage of the pipeline
                No additional parameters may be passed to this prompt template.
            base_model_kwargs: Any additional keyword arguments that should be passed to the
                LLMClient.query_model call for the normal stages of the pipeline.
            conversion_model_kwargs: Any additional keyword arguments that should be passed to the
                LLMClient.query_model call for the yaml conversion stages of the pipeline.
            ignore_conversion_failure: Ignores yaml conversion failures when able and discards the data
                that conversion was attempted on
        Returns:
            A list of synthetically generated writing task prompts
        """
        self.logger.info("Starting writing pipeline")
        # Generate the tasks
        raw_writing_tasks = []
        for topic in topics:
            for material in text_material_types:
                raw_writing_tasks.append(
                    self._generate_parse_writing_task(
                        topic=topic,
                        material=material,
                        n_openlines=n_openlines,
                        model=model,
                        writing_task_prompt_template=writing_task_prompt_template,
                        yaml_conversion_prompt_template=yaml_conversion_prompt_template,
                        base_model_kwargs=base_model_kwargs,
                        conversion_model_kwargs=conversion_model_kwargs,
                        ignore_conversion_failure=ignore_conversion_failure,
                    )
                )
        self.logger.info("Starting writing task generation")
        raw_writing_tasks = await self._gather(raw_writing_tasks)
        writing_tasks = [item for tasks in raw_writing_tasks for item in tasks]
        self.logger.info("Finished writing task generation")

        # Revise the tasks
        raw_revised_openlines = [
            self._revise_parse_writing_task(
                task=task,
                n_revisions=n_revisions,
                model=model,
                revise_writing_task_prompt_template=revise_writing_task_prompt_template,
                yaml_conversion_prompt_template=yaml_conversion_prompt_template,
                base_model_kwargs=base_model_kwargs,
                conversion_model_kwargs=conversion_model_kwargs,
                ignore_conversion_failure=ignore_conversion_failure,
            )
            for task in writing_tasks
        ]
        self.logger.info("Starting writing task revision")
        raw_revised_openlines = await self._gather(raw_revised_openlines)
        revised_openlines = [item for lines in raw_revised_openlines for item in lines]
        self.logger.info("Finished writing task revision")
        self.logger.info("Finished writing pipeline")

        return revised_openlines

    async def _generate_parse_writing_task(
        self,
        topic: str,
        material: str,
        n_openlines: Union[int, str],
        model: str,
        writing_task_prompt_template: str,
        yaml_conversion_prompt_template: str,
        base_model_kwargs: dict,
        conversion_model_kwargs: dict,
        ignore_conversion_failure: bool,
    ) -> List[str]:
        raw_tasks = await self.generate_writing_tasks(
            topic=topic,
            text_material_type=material,
            n_openlines=n_openlines,
            model=model,
            model_kwargs=base_model_kwargs,
            prompt_template=writing_task_prompt_template,
        )
        raw_tasks = raw_tasks[0]
        return await self._try_convert_yaml_list(
            raw_tasks,
            model=model,
            yaml_conversion_prompt_template=yaml_conversion_prompt_template,
            conversion_model_kwargs=conversion_model_kwargs,
            expected_length=n_openlines,
            ignore_conversion_failure=ignore_conversion_failure,
        )

    async def _revise_parse_writing_task(
        self,
        task: str,
        n_revisions: Union[int, str],
        model: str,
        revise_writing_task_prompt_template: str,
        yaml_conversion_prompt_template: str,
        base_model_kwargs: dict,
        conversion_model_kwargs: dict,
        ignore_conversion_failure: bool,
    ) -> List[str]:
        raw_revision = await self.revise_writing_tasks(
            openline=task,
            n_revisions=n_revisions,
            model=model,
            model_kwargs=base_model_kwargs,
            prompt_template=revise_writing_task_prompt_template,
        )
        raw_revision = raw_revision[0]
        return await self._try_convert_yaml_list(
            raw_revision,
            model=model,
            yaml_conversion_prompt_template=yaml_conversion_prompt_template,
            conversion_model_kwargs=conversion_model_kwargs,
            expected_length=n_revisions,
            ignore_conversion_failure=ignore_conversion_failure,
        )

    async def run_closed_qa_pipeline(
        self,
        documents: List[str],
        n_openlines: Union[str, int],
        model: str,
        closed_qa_prompt_template: str = DEFAULT_CLOSED_QA_PROMPT_TEMPLATE,
        yaml_conversion_prompt_template: str = DEFAULT_YAML_CONVERSION_PROMPT_TEMPLATE,
        base_model_kwargs: dict = {},
        conversion_model_kwargs: dict = {},
        ignore_conversion_failure: bool = False,
    ) -> List[Tuple[int, str]]:
        """
        Runs a pipeline for automatically generating closed Q&A openlines for a dialogue
        Args:
            documents: A list of documents to generate closed Q&A questions for
            n_openlines: The number of questions to generate per document.
            model: The name of the model that should be used to generate all the responses.
                Must be available in the LLMClient passed in the constructor.
            closed_qa_prompt_template: A format string of the prompt to use. It must have the following parameters:
                - n_openlines: Will be populated with the n_openlines passed in this function
                - document: Will be populated with one element of the documents list passed in this function
                No additional parameters may be passed to this prompt template.
            yaml_conversion_prompt_template: A format string of the prompt to use. It must have the following parameters:
                - llm_response: Will be populated with the raw LLM response from each stage of the pipeline
                No additional parameters may be passed to this prompt template.
            base_model_kwargs: Any additional keyword arguments that should be passed to the
                LLMClient.query_model call for the normal stages of the pipeline.
            conversion_model_kwargs: Any additional keyword arguments that should be passed to the
                LLMClient.query_model call for the yaml conversion stages of the pipeline.
            ignore_conversion_failure: Ignores yaml conversion failures when able and discards the data
                that conversion was attempted on
        Returns:
            A list of pairs where the first element represents the index of the document used to generate the question in the documents list
            and the second element represents a synthetically generated closed Q&A prompt. Example: [(0, "Summarize this document"), ...]
        """
        self.logger.info("Starting closed q&a pipeline")
        raw_qa = [
            self._generate_parse_closed_qa(
                document_id=i,
                document=document,
                n_openlines=n_openlines,
                model=model,
                closed_qa_prompt_template=closed_qa_prompt_template,
                yaml_conversion_prompt_template=yaml_conversion_prompt_template,
                base_model_kwargs=base_model_kwargs,
                conversion_model_kwargs=conversion_model_kwargs,
                ignore_conversion_failure=ignore_conversion_failure,
            )
            for i, document in enumerate(documents)
        ]
        raw_qa = await self._gather(raw_qa)
        document_openline_pairs = [item for lines in raw_qa for item in lines]
        self.logger.info("Finished closed q&a pipeline")

        return document_openline_pairs

    async def _generate_parse_closed_qa(
        self,
        document_id: int,
        document: str,
        n_openlines: Union[int, str],
        model: str,
        closed_qa_prompt_template: str,
        yaml_conversion_prompt_template: str,
        base_model_kwargs: dict,
        conversion_model_kwargs: dict,
        ignore_conversion_failure: bool,
    ) -> List[str]:
        raw_instruction = await self.generate_closed_qa_instructions(
            document=document,
            n_openlines=n_openlines,
            model=model,
            model_kwargs=base_model_kwargs,
            prompt_template=closed_qa_prompt_template,
        )
        raw_instruction = raw_instruction[0]
        parsed_instructions = await self._try_convert_yaml_list(
            raw_instruction,
            model=model,
            yaml_conversion_prompt_template=yaml_conversion_prompt_template,
            conversion_model_kwargs=conversion_model_kwargs,
            expected_length=n_openlines,
            ignore_conversion_failure=ignore_conversion_failure,
        )

        return [(document_id, inst) for inst in parsed_instructions]

    async def run_math_pipeline(
        self,
        n_macro_topics: Union[str, int],
        school_level: str,
        n_subtopics: Union[str, int],
        n_openlines: Union[str, int],
        model: str,
        macro_topic_prompt_template: str = DEFAULT_MATH_MACRO_TOPICS_PROMPT_TEMPLATE,
        subtopic_prompt_template: str = DEFAULT_MATH_SUBTOPICS_PROMPT_TEMPLATE,
        math_problem_prompt_template: str = MATH_PROBLEM_GENERAL_PROMPT_TEMPLATE,
        yaml_conversion_prompt_template: str = DEFAULT_YAML_CONVERSION_PROMPT_TEMPLATE,
        base_model_kwargs: dict = {},
        conversion_model_kwargs: dict = {},
        additional_macro_topics: List[str] = [],
        additional_subtopics: List[str] = [],
        ignore_conversion_failure: bool = False,
        combine_topics: bool = True,
    ) -> List[str]:
        """
        Runs a pipeline for automatically generating math questions for a dialogue
        Args:
            n_macro_topics: The number of macro topics to generate.
            school_level: The school level to target when generating macro topics.
            n_subtopics: The number of subtopics to generate per macro topic.
            n_openlines: The number of questions to generate per topic.
            model: The name of the model that should be used to generate all the responses.
                Must be available in the LLMClient passed in the constructor.
            macro_topic_prompt_template: A format string of the prompt to use. It must have the following parameters:
                - n_macro_topics: Will be populated with the n_macro_topics passed in this function
                - school_level: Will be populated with the school_level passed in this function
                No additional parameters may be passed to this prompt template.
            subtopic_prompt_template: A format string of the prompt to use. It must have the following parameters:
                - n_subtopics: Will be populated with the n_subtopics passed in this function
                - macro_topic: Will be populated with a generated macro topic
                No additional parameters may be passed to this prompt template.
            math_problem_prompt_template: A format string of the prompt to use. It must have the following parameters:
                - n_openlines: Will be populated with the n_openlines passed in this function
                - topic: Will be populated with a generated topic
                No additional parameters may be passed to this prompt template.
                Some example templates found in nemo_curator.synthetic include:
                - MATH_PROBLEM_GENERAL_PROMPT_TEMPLATE
                - MATH_PROBLEM_BEGINNER_PROMPT_TEMPLATE
            yaml_conversion_prompt_template: A format string of the prompt to use. It must have the following parameters:
                - llm_response: Will be populated with the raw LLM response from each stage of the pipeline
                No additional parameters may be passed to this prompt template.
            base_model_kwargs: Any additional keyword arguments that should be passed to the
                LLMClient.query_model call for the normal stages of the pipeline.
            conversion_model_kwargs: Any additional keyword arguments that should be passed to the
                LLMClient.query_model call for the yaml conversion stages of the pipeline.
            ignore_conversion_failure: Ignores yaml conversion failures when able and discards the data
                that conversion was attempted on
            combine_topics: If True, mixes the macro topics with the subtopics when generating openlines.
                If False, only the subtopics are used.
        Returns:
            A list of synthetically generated math prompts
        """
        self.logger.info("Starting math pipeline")
        # Generate the macro topics
        self.logger.info("Starting math macro topic generation")
        responses = await self.generate_math_macro_topics(
            n_macro_topics=n_macro_topics,
            school_level=school_level,
            model=model,
            model_kwargs=base_model_kwargs,
            prompt_template=macro_topic_prompt_template,
        )
        try:
            macro_topics = await self.convert_response_to_yaml_list(
                responses[0],
                model=model,
                prompt_template=yaml_conversion_prompt_template,
                model_kwargs=conversion_model_kwargs,
            )
            if len(macro_topics) != n_macro_topics and not ignore_conversion_failure:
                raise YamlConversionError(
                    f"Error: Length of macro topics {len(macro_topics)} does not match desired n_macro_topics {n_macro_topics}: {macro_topics}"
                )
        except YamlConversionError as e:
            if ignore_conversion_failure:
                macro_topics = []
            else:
                raise e
        macro_topics.extend(additional_macro_topics)
        self.logger.info("Finished math macro topic generation")

        # Generate the subtopics
        raw_topics = [
            self._generate_parse_math_subtopic(
                macro_topic=macro_topic,
                n_subtopics=n_subtopics,
                model=model,
                subtopic_prompt_template=subtopic_prompt_template,
                yaml_conversion_prompt_template=yaml_conversion_prompt_template,
                base_model_kwargs=base_model_kwargs,
                conversion_model_kwargs=conversion_model_kwargs,
                ignore_conversion_failure=ignore_conversion_failure,
            )
            for macro_topic in macro_topics
        ]
        self.logger.info("Starting math subtopic generation")
        raw_topics = await self._gather(raw_topics)
        topic_list = [item for subtopics in raw_topics for item in subtopics]
        topic_list.extend(additional_subtopics)
        self.logger.info("Finished math subtopic generation")

        # Mix the macro topics with the subtopics
        if combine_topics:
            topic_list.extend(macro_topics)

        # Generate the openlines
        raw_lines = [
            self._generate_parse_math_openline(
                subtopic=subtopic,
                n_openlines=n_openlines,
                model=model,
                math_problem_prompt_template=math_problem_prompt_template,
                yaml_conversion_prompt_template=yaml_conversion_prompt_template,
                base_model_kwargs=base_model_kwargs,
                conversion_model_kwargs=conversion_model_kwargs,
                ignore_conversion_failure=ignore_conversion_failure,
            )
            for subtopic in topic_list
        ]
        self.logger.info("Starting math openline generation")
        raw_lines = await self._gather(raw_lines)
        openlines = [item for lines in raw_lines for item in lines]
        self.logger.info("Finished math openline generation")
        self.logger.info("Finished math pipeline")

        return openlines

    async def _generate_parse_math_subtopic(
        self,
        macro_topic: str,
        n_subtopics: Union[int, str],
        model: str,
        subtopic_prompt_template: str,
        yaml_conversion_prompt_template: str,
        base_model_kwargs: dict,
        conversion_model_kwargs: dict,
        ignore_conversion_failure: bool,
    ) -> List[str]:
        raw_topic = await self.generate_math_subtopics(
            macro_topic=macro_topic,
            n_subtopics=n_subtopics,
            model=model,
            model_kwargs=base_model_kwargs,
            prompt_template=subtopic_prompt_template,
        )
        raw_topic = raw_topic[0]
        return await self._try_convert_yaml_list(
            raw_topic,
            model=model,
            yaml_conversion_prompt_template=yaml_conversion_prompt_template,
            conversion_model_kwargs=conversion_model_kwargs,
            expected_length=n_subtopics,
            ignore_conversion_failure=ignore_conversion_failure,
        )

    async def _generate_parse_math_openline(
        self,
        subtopic: str,
        n_openlines: Union[int, str],
        model: str,
        math_problem_prompt_template: str,
        yaml_conversion_prompt_template: str,
        base_model_kwargs: dict,
        conversion_model_kwargs: dict,
        ignore_conversion_failure: bool,
    ) -> List[str]:
        raw_line = await self.generate_math_problem(
            topic=subtopic,
            n_openlines=n_openlines,
            model=model,
            model_kwargs=base_model_kwargs,
            prompt_template=math_problem_prompt_template,
        )
        raw_line = raw_line[0]
        return await self._try_convert_yaml_list(
            raw_line,
            model=model,
            yaml_conversion_prompt_template=yaml_conversion_prompt_template,
            conversion_model_kwargs=conversion_model_kwargs,
            expected_length=n_openlines,
            ignore_conversion_failure=ignore_conversion_failure,
        )

    async def run_python_pipeline(
        self,
        n_macro_topics: Union[str, int],
        n_subtopics: Union[str, int],
        n_openlines: Union[str, int],
        model: str,
        macro_topic_prompt_template: str = DEFAULT_PYTHON_MACRO_TOPICS_PROMPT_TEMPLATE,
        subtopic_prompt_template: str = DEFAULT_PYTHON_SUBTOPICS_PROMPT_TEMPLATE,
        python_problem_prompt_template: str = PYTHON_PROBLEM_BEGINNER_PROMPT_TEMPLATE,
        yaml_conversion_prompt_template: str = DEFAULT_YAML_CONVERSION_PROMPT_TEMPLATE,
        base_model_kwargs: dict = {},
        conversion_model_kwargs: dict = {},
        additional_macro_topics: List[str] = [],
        additional_subtopics: List[str] = [],
        ignore_conversion_failure: bool = False,
        combine_topics: bool = True,
    ) -> List[str]:
        """
        Runs a pipeline for automatically generating Python questions for a dialogue
        Args:
            n_macro_topics: The number of macro topics to generate.
            n_subtopics: The number of subtopics to generate per macro topic.
            n_openlines: The number of questions to generate per topic.
            model: The name of the model that should be used to generate all the responses.
                Must be available in the LLMClient passed in the constructor.
            macro_topic_prompt_template: A format string of the prompt to use. It must have the following parameters:
                - n_macro_topics: Will be populated with the n_macro_topics passed in this function
                No additional parameters may be passed to this prompt template.
            subtopic_prompt_template: A format string of the prompt to use. It must have the following parameters:
                - n_subtopics: Will be populated with the n_subtopics passed in this function
                - macro_topic: Will be populated with a generated macro topic
                No additional parameters may be passed to this prompt template.
            python_problem_prompt_template: A format string of the prompt to use. It must have the following parameters:
                - n_openlines: Will be populated with the n_openlines passed in this function
                - language: Will be populated with "Python"
                - topic: Will be populated with a generated topic
                No additional parameters may be passed to this prompt template.
                Some example templates found in nemo_curator.synthetic include:
                - PYTHON_PROBLEM_BEGINNER_PROMPT_TEMPLATE
                - PYTHON_PROBLEM_INTERMEDIATE_PROMPT_TEMPLATE
                - PYTHON_PROBLEM_ADVANCED_PROMPT_TEMPLATE
            yaml_conversion_prompt_template: A format string of the prompt to use. It must have the following parameters:
                - llm_response: Will be populated with the raw LLM response from each stage of the pipeline
                No additional parameters may be passed to this prompt template.
            base_model_kwargs: Any additional keyword arguments that should be passed to the
                LLMClient.query_model call for the normal stages of the pipeline.
            conversion_model_kwargs: Any additional keyword arguments that should be passed to the
                LLMClient.query_model call for the yaml conversion stages of the pipeline.
            ignore_conversion_failure: Ignores yaml conversion failures when able and discards the data
                that conversion was attempted on
            combine_topics: If True, mixes the macro topics with the subtopics when generating openlines.
                If False, only the subtopics are used.
        Returns:
            A list of synthetically generated Python prompts
        """
        self.logger.info("Starting python pipeline")
        # Generate the macro topics
        self.logger.info("Starting python macro topic generation")
        responses = await self.generate_python_macro_topics(
            n_macro_topics=n_macro_topics,
            model=model,
            model_kwargs=base_model_kwargs,
            prompt_template=macro_topic_prompt_template,
        )
        try:
            macro_topics = await self.convert_response_to_yaml_list(
                responses[0],
                model=model,
                prompt_template=yaml_conversion_prompt_template,
                model_kwargs=conversion_model_kwargs,
            )
            if len(macro_topics) != n_macro_topics and not ignore_conversion_failure:
                raise YamlConversionError(
                    f"Error: Length of macro topics {len(macro_topics)} does not match desired n_macro_topics {n_macro_topics}: {macro_topics}"
                )
        except YamlConversionError as e:
            if ignore_conversion_failure:
                macro_topics = []
            else:
                raise e
        macro_topics.extend(additional_macro_topics)
        self.logger.info("Finished python macro topic generation")

        # Generate the subtopics
        raw_topics = [
            self._generate_parse_python_subtopic(
                macro_topic=macro_topic,
                n_subtopics=n_subtopics,
                model=model,
                subtopic_prompt_template=subtopic_prompt_template,
                yaml_conversion_prompt_template=yaml_conversion_prompt_template,
                base_model_kwargs=base_model_kwargs,
                conversion_model_kwargs=conversion_model_kwargs,
                ignore_conversion_failure=ignore_conversion_failure,
            )
            for macro_topic in macro_topics
        ]
        self.logger.info("Starting python subtopic generation")
        raw_topics = await self._gather(raw_topics)
        topic_list = [item for subtopics in raw_topics for item in subtopics]
        topic_list.extend(additional_subtopics)
        self.logger.info("Finished python subtopic generation")

        # Mix the macro topics with the subtopics
        if combine_topics:
            topic_list.extend(macro_topics)

        # Generate the openlines
        raw_lines = [
            self._generate_parse_python_openline(
                subtopic=subtopic,
                n_openlines=n_openlines,
                model=model,
                python_problem_prompt_template=python_problem_prompt_template,
                yaml_conversion_prompt_template=yaml_conversion_prompt_template,
                base_model_kwargs=base_model_kwargs,
                conversion_model_kwargs=conversion_model_kwargs,
                ignore_conversion_failure=ignore_conversion_failure,
            )
            for subtopic in topic_list
        ]
        self.logger.info("Starting python openline generation")
        raw_lines = await self._gather(raw_lines)
        openlines = [item for lines in raw_lines for item in lines]
        self.logger.info("Finished python openline generation")
        self.logger.info("Finished python pipeline")

        return openlines

    async def _generate_parse_python_subtopic(
        self,
        macro_topic: str,
        n_subtopics: Union[int, str],
        model: str,
        subtopic_prompt_template: str,
        yaml_conversion_prompt_template: str,
        base_model_kwargs: dict,
        conversion_model_kwargs: dict,
        ignore_conversion_failure: bool,
    ) -> List[str]:
        raw_topic = await self.generate_python_subtopics(
            macro_topic=macro_topic,
            n_subtopics=n_subtopics,
            model=model,
            model_kwargs=base_model_kwargs,
            prompt_template=subtopic_prompt_template,
        )
        raw_topic = raw_topic[0]
        return await self._try_convert_yaml_list(
            raw_topic,
            model=model,
            yaml_conversion_prompt_template=yaml_conversion_prompt_template,
            conversion_model_kwargs=conversion_model_kwargs,
            expected_length=n_subtopics,
            ignore_conversion_failure=ignore_conversion_failure,
        )

    async def _generate_parse_python_openline(
        self,
        subtopic: str,
        n_openlines: Union[int, str],
        model: str,
        python_problem_prompt_template: str,
        yaml_conversion_prompt_template: str,
        base_model_kwargs: dict,
        conversion_model_kwargs: dict,
        ignore_conversion_failure: bool,
    ) -> List[str]:
        raw_line = await self.generate_python_problem(
            topic=subtopic,
            n_openlines=n_openlines,
            model=model,
            model_kwargs=base_model_kwargs,
            prompt_template=python_problem_prompt_template,
        )
        raw_line = raw_line[0]
        return await self._try_convert_yaml_list(
            raw_line,
            model=model,
            yaml_conversion_prompt_template=yaml_conversion_prompt_template,
            conversion_model_kwargs=conversion_model_kwargs,
            expected_length=n_openlines,
            ignore_conversion_failure=ignore_conversion_failure,
        )



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from unittest.mock import MagicMock, patch

import pytest

from nemo_curator.nemo_run.slurm import SlurmJobConfig


class TestSlurmJobConfig:

    @pytest.fixture
    def basic_config(self):
        """Returns a basic SlurmJobConfig with required parameters"""
        return SlurmJobConfig(
            job_dir="/path/to/job",
            container_entrypoint="/path/to/container-entrypoint.sh",
            script_command="nemo_curator_tool arg1 arg2",
        )

    @pytest.fixture
    def custom_config(self):
        """Returns a SlurmJobConfig with custom parameters"""
        return SlurmJobConfig(
            job_dir="/path/to/custom/job",
            container_entrypoint="/path/to/custom/container-entrypoint.sh",
            script_command="custom_tool --arg1=val1 --arg2=val2",
            device="gpu",
            interface="ib0",
            protocol="ucx",
            cpu_worker_memory_limit="8GB",
            rapids_no_initialize="0",
            cudf_spill="0",
            rmm_scheduler_pool_size="2GB",
            rmm_worker_pool_size="80GiB",
            libcudf_cufile_policy="ON",
        )

    def test_initialize_with_defaults(self, basic_config):
        """Test initializing SlurmJobConfig with default values"""
        # Check required parameters
        assert basic_config.job_dir == "/path/to/job"
        assert basic_config.container_entrypoint == "/path/to/container-entrypoint.sh"
        assert basic_config.script_command == "nemo_curator_tool arg1 arg2"

        # Check default values
        assert basic_config.device == "cpu"
        assert basic_config.interface == "eth0"
        assert basic_config.protocol == "tcp"
        assert basic_config.cpu_worker_memory_limit == "0"
        assert basic_config.rapids_no_initialize == "1"
        assert basic_config.cudf_spill == "1"
        assert basic_config.rmm_scheduler_pool_size == "1GB"
        assert basic_config.rmm_worker_pool_size == "72GiB"
        assert basic_config.libcudf_cufile_policy == "OFF"

    def test_initialize_with_custom_values(self, custom_config):
        """Test initializing SlurmJobConfig with custom values"""
        # Check required parameters
        assert custom_config.job_dir == "/path/to/custom/job"
        assert (
            custom_config.container_entrypoint
            == "/path/to/custom/container-entrypoint.sh"
        )
        assert custom_config.script_command == "custom_tool --arg1=val1 --arg2=val2"

        # Check custom values
        assert custom_config.device == "gpu"
        assert custom_config.interface == "ib0"
        assert custom_config.protocol == "ucx"
        assert custom_config.cpu_worker_memory_limit == "8GB"
        assert custom_config.rapids_no_initialize == "0"
        assert custom_config.cudf_spill == "0"
        assert custom_config.rmm_scheduler_pool_size == "2GB"
        assert custom_config.rmm_worker_pool_size == "80GiB"
        assert custom_config.libcudf_cufile_policy == "ON"

    def test_build_env_vars(self, basic_config):
        """Test the _build_env_vars method"""
        env_vars = basic_config._build_env_vars()

        # Check that keys are uppercased
        assert "JOB_DIR" in env_vars
        assert "CONTAINER_ENTRYPOINT" in env_vars
        assert "SCRIPT_COMMAND" in env_vars
        assert "DEVICE" in env_vars

        # Check derived variables
        assert env_vars["LOGDIR"] == "/path/to/job/logs"
        assert env_vars["PROFILESDIR"] == "/path/to/job/profiles"
        assert env_vars["SCHEDULER_FILE"] == "/path/to/job/logs/scheduler.json"
        assert env_vars["SCHEDULER_LOG"] == "/path/to/job/logs/scheduler.log"
        assert env_vars["DONE_MARKER"] == "/path/to/job/logs/done.txt"

        # Check values
        assert env_vars["JOB_DIR"] == "/path/to/job"
        assert env_vars["CONTAINER_ENTRYPOINT"] == "/path/to/container-entrypoint.sh"
        assert env_vars["SCRIPT_COMMAND"] == "nemo_curator_tool arg1 arg2"
        assert env_vars["DEVICE"] == "cpu"

    def test_to_script_with_defaults(self, basic_config):
        """Test to_script method with default arguments"""
        # Mock the run.Script class
        with patch("nemo_curator.nemo_run.slurm.run") as mock_run:
            mock_script = MagicMock()
            mock_run.Script.return_value = mock_script

            script = basic_config.to_script()

            # Verify Script was created with correct parameters
            mock_run.Script.assert_called_once()
            call_args = mock_run.Script.call_args[1]
            assert call_args["path"] == "/path/to/container-entrypoint.sh"

            # Check env variables
            env_vars = call_args["env"]
            assert (
                env_vars["SCRIPT_COMMAND"]
                == '"nemo_curator_tool arg1 arg2 --scheduler-file=/path/to/job/logs/scheduler.json --device=cpu"'
            )

            # Check that the script object was returned
            assert script == mock_script

    def test_to_script_with_custom_options(self, basic_config):
        """Test to_script method with custom add_scheduler_file and add_device arguments"""
        # Mock the run.Script class
        with patch("nemo_curator.nemo_run.slurm.run") as mock_run:
            mock_script = MagicMock()
            mock_run.Script.return_value = mock_script

            # Don't add scheduler file or device arguments
            script = basic_config.to_script(add_scheduler_file=False, add_device=False)

            # Verify Script was created with correct parameters
            mock_run.Script.assert_called_once()
            call_args = mock_run.Script.call_args[1]
            assert call_args["path"] == "/path/to/container-entrypoint.sh"

            # Check env variables - should not contain scheduler file or device
            env_vars = call_args["env"]
            assert env_vars["SCRIPT_COMMAND"] == '"nemo_curator_tool arg1 arg2"'

            # Check that it doesn't have the scheduler file or device arguments
            assert "--scheduler-file" not in env_vars["SCRIPT_COMMAND"]
            assert "--device" not in env_vars["SCRIPT_COMMAND"]

    def test_to_script_with_gpu_device(self, custom_config):
        """Test to_script method with a GPU device configuration"""
        # Mock the run.Script class
        with patch("nemo_curator.nemo_run.slurm.run") as mock_run:
            mock_script = MagicMock()
            mock_run.Script.return_value = mock_script

            script = custom_config.to_script()

            # Verify Script was created with correct parameters
            call_args = mock_run.Script.call_args[1]
            env_vars = call_args["env"]

            # Check that the device is set to GPU
            assert "--device=gpu" in env_vars["SCRIPT_COMMAND"]
            assert env_vars["DEVICE"] == "gpu"

            # Check UCX-specific settings
            assert env_vars["PROTOCOL"] == "ucx"
            assert env_vars["RMM_WORKER_POOL_SIZE"] == "80GiB"



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse

import nemo_curator as nc
from nemo_curator.datasets import DocumentDataset
from nemo_curator.utils.distributed_utils import get_client
from nemo_curator.utils.script_utils import ArgumentHelper


def main(args):
    # Params
    dataset_paths = ["/path/to/first", "/path/to/second", "/path/to/third"]
    dataset_weights = [5.0, 2.0, 1.0]
    target_size = 1000
    output_path = "/path/to/output"

    # Set up Dask client
    client = get_client(**ArgumentHelper.parse_client_args(args))

    # Blend the datasets
    datasets = [DocumentDataset.read_json(path) for path in dataset_paths]
    blended_dataset = nc.blend_datasets(target_size, datasets, dataset_weights)

    shuffle = nc.Shuffle(seed=42)
    blended_dataset = shuffle(blended_dataset)

    # Save the blend
    blended_dataset.to_json(output_path)


def attach_args(
    parser=argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    ),
):
    return ArgumentHelper(parser).add_distributed_args()


if __name__ == "__main__":
    main(attach_args().parse_args())



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import argparse
import os

import psutil

from nemo_curator import __version__


class ArgumentHelper:
    """
    A helper class to add common arguments to an argparse.ArgumentParser instance.
    """

    def __init__(self, parser: argparse.ArgumentParser):
        self.parser = parser

        version_string = f"NVIDIA NeMo Curator -- v{__version__}"
        self.attach_version_arg(version_string)
        parser_print_help = self.parser.print_help

        def print_help_with_version(*args, **kwargs):
            print(version_string)
            parser_print_help(*args, **kwargs)

        self.parser.print_help = print_help_with_version

    def attach_version_arg(self, version_string: str):
        self.parser.add_argument(
            "--version",
            "-v",
            action="version",
            version=version_string,
            help="Show the version and exit.",
        )

    @staticmethod
    def attach_bool_arg(
        parser: argparse.ArgumentParser,
        flag_name: str,
        default: bool = False,
        help: str = None,
    ):
        attr_name = flag_name.replace("-", "_")
        help = flag_name.replace("-", " ") if help is None else help
        parser.add_argument(
            "--{}".format(flag_name),
            dest=attr_name,
            action="store_true",
            help=help,
        )
        parser.add_argument(
            "--no-{}".format(flag_name),
            dest=attr_name,
            action="store_false",
            help=help,
        )

        parser.set_defaults(**{attr_name: default})

    def add_arg_batch_size(
        self,
        default: int = 64,
        help: str = "Number of files to read into memory at a time.",
    ):
        self.parser.add_argument(
            "--batch-size",
            type=int,
            default=default,
            help=help,
        )

    def add_arg_device(self):
        self.parser.add_argument(
            "--device",
            type=str,
            default="gpu",
            help='Device to run the script on. Either "cpu" or "gpu".',
        )

    def add_arg_enable_spilling(self):
        self.parser.add_argument("--enable-spilling", action="store_true")

    def add_arg_language(self, help: str):
        self.parser.add_argument(
            "--language",
            type=str,
            default="en",
            help=help,
        )

    def add_arg_log_dir(self, default: str):
        self.parser.add_argument(
            "--log-dir",
            type=str,
            default=default,
            help="The output log directory where node and local"
            " ranks will write their respective log files.",
        )

    def add_arg_input_data_dir(
        self,
        required=False,
        help: str = "Input directory consisting of .jsonl files that are accessible "
        "to all nodes. Use this for a distributed file system.",
    ):
        self.parser.add_argument(
            "--input-data-dir",
            type=str,
            default=None,
            required=required,
            help=help,
        )

    def add_arg_input_file_type(
        self,
        choices=None,
        required=False,
        help="File type of the dataset to be read in. Supported file formats "
        'include "jsonl" (default), "pickle", or "parquet".',
    ):
        self.parser.add_argument(
            "--input-file-type",
            type=str,
            default="jsonl",
            required=required,
            choices=choices,
            help=help,
        )

    def add_arg_input_file_extension(
        self,
        help: str = "The file extension of the input files. If not provided, the input file type will be used.",
    ):
        self.parser.add_argument(
            "--input-file-extension",
            type=str,
            default=None,
            help=help,
        )

    def add_arg_input_local_data_dir(self):
        self.parser.add_argument(
            "--input-local-data-dir",
            type=str,
            default=None,
            help="Input directory consisting of dataset files. "
            "Use this argument when a distributed file system is not available.",
        )

    def add_arg_input_meta(self):
        self.parser.add_argument(
            "--input-meta",
            type=str,
            default=None,
            help="A string formatted as a dictionary, which outlines the field names and "
            "their respective data types within the JSONL input files.",
        )

    def add_arg_input_text_field(self):
        self.parser.add_argument(
            "--input-text-field",
            type=str,
            default="text",
            help="The name of the field within each datapoint object of the input "
            "file that contains the text.",
        )

    def add_arg_id_column(self):
        self.parser.add_argument(
            "--id-column",
            type=str,
            default="id",
            help="The name of the field within each datapoint object of the input "
            "file that contains the ID.",
        )

    def add_arg_id_column_type(self):
        self.parser.add_argument(
            "--id-column-type",
            type=str,
            default="int",
            help='The datatype of the ID field, either "int" or "str".',
        )

    def add_arg_minhash_length(self):
        self.parser.add_argument(
            "--minhash-length",
            type=int,
            default=260,
            help="The minhash signature length of each input document.",
        )

    def add_arg_nvlink_only(self):
        self.parser.add_argument(
            "--nvlink-only",
            action="store_true",
            help="Start a local cluster with only NVLink enabled."
            'Only applicable when protocol="ucx" and no scheduler file or address is specified.',
        )

    def add_arg_output_data_dir(self, help: str):
        self.parser.add_argument(
            "--output-data-dir",
            type=str,
            required=True,
            help=help,
        )

    def add_arg_output_dir(
        self, required=False, help: str = "The output directory to write results."
    ):
        self.parser.add_argument(
            "--output-dir",
            type=str,
            required=required,
            help=help,
        )

    def add_arg_output_file_type(
        self,
        choices=None,
        help="File type the dataset will be written to. Supported file formats "
        'include "jsonl" (default), "pickle", or "parquet".',
    ):
        self.parser.add_argument(
            "--output-file-type",
            type=str,
            default="jsonl",
            choices=choices,
            help=help,
        )

    def add_arg_output_train_file(self, help: str, default: str = None):
        self.parser.add_argument(
            "--output-train-file",
            type=str,
            default=default,
            help=help,
        )

    def add_arg_protocol(self):
        self.parser.add_argument(
            "--protocol",
            type=str,
            default="ucx",
            help="Protocol to use for Dask cluster. "
            "Note: This only applies to the LocalCUDACluster. If providing a user-created "
            "cluster, please refer to "
            "https://docs.rapids.ai/api/dask-cuda/stable/api.html#cmdoption-dask-cuda-protocol.",  # noqa: E501
        )

    def add_arg_rmm_pool_size(self):
        self.parser.add_argument(
            "--rmm-pool-size",
            type=str,
            default="14GB",
            help="Initial pool size to use for the RMM Pool Memory allocator. "
            "Note: This only applies to the LocalCUDACluster. If providing a user-created "
            "cluster, please refer to "
            "https://docs.rapids.ai/api/dask-cuda/stable/api.html#cmdoption-dask-cuda-rmm-pool-size.",  # noqa: E501
        )

    def add_arg_scheduler_address(self):
        self.parser.add_argument(
            "--scheduler-address",
            type=str,
            default=None,
            help="Address to the scheduler of a created Dask cluster. If not provided, "
            "a single-node LocalCUDACluster will be started.",
        )

    def add_arg_scheduler_file(self):
        self.parser.add_argument(
            "--scheduler-file",
            type=str,
            default=None,
            help="Path to the scheduler file of a created Dask cluster. If not provided, "
            " a single-node LocalCUDACluster will be started.",
        )

    def add_arg_seed(
        self,
        default=42,
        help: str = "If specified, the random seed used for shuffling.",
    ):
        self.parser.add_argument(
            "--seed",
            type=int,
            default=default,
            help=help,
        )

    def add_arg_set_torch_to_use_rmm(self):
        self.parser.add_argument("--set-torch-to-use-rmm", action="store_true")

    def add_arg_shuffle(self, help: str):
        ArgumentHelper.attach_bool_arg(
            self.parser,
            "shuffle",
            help=help,
        )

    def add_arg_text_ddf_blocksize(self):
        self.parser.add_argument(
            "--text-ddf-blocksize",
            type=int,
            default=256,
            help="The block size for chunking JSONL files for text DataFrames in MB.",
        )

    def add_arg_model_path(self, help="The path to the model file"):
        self.parser.add_argument(
            "--pretrained-model-name-or-path",
            type=str,
            help=help,
            required=False,
        )

    def add_arg_max_mem_gb_classifier(self):
        self.parser.add_argument(
            "--max-mem-gb-classifier",
            default=None,
            type=int,
            help="Specify the maximum GPU memory (in GB) for the classifier. "
            "Defaults to using the total GPU memory minus 4 GB if not specified.",
        )

    def add_arg_autocast(self, help="Whether to use autocast or not"):
        ArgumentHelper.attach_bool_arg(
            parser=self.parser,
            flag_name="autocast",
            default=True,
            help=help,
        )

    def add_arg_max_chars(self, default=2000):
        self.parser.add_argument(
            "--max-chars",
            type=int,
            default=default,
            help="Truncates all documents in the dataset to this number of characters before running model inference on them.",
        )

    def add_distributed_args(self) -> argparse.ArgumentParser:
        """
        Adds default set of arguments that are needed for Dask cluster setup.

        """
        self.parser.add_argument(
            "--device",
            type=str,
            default="cpu",
            help='Device to run the script on. Either "cpu" or "gpu".',
        )
        self.parser.add_argument(
            "--files-per-partition",
            type=int,
            default=2,
            help="Number of JSONL files to combine into a single partition.",
        )
        self.parser.add_argument(
            "--n-workers",
            type=int,
            default=os.cpu_count(),
            help="The number of workers to run in total on the Dask CPU cluster.",
        )
        self.parser.add_argument(
            "--num-files",
            type=int,
            default=None,
            help="Upper limit on the number of JSON files to process.",
        )
        self.parser.add_argument(
            "--nvlink-only",
            action="store_true",
            help="Start a local cluster with only NVLink enabled. "
            'Only applicable when protocol="ucx" and no scheduler file or address is specified.',
        )
        self.parser.add_argument(
            "--protocol",
            type=str,
            default="tcp",
            help="Protcol to use for Dask cluster"
            "Note: This only applies to the LocalCUDACluster. If providing a user-created "
            "cluster, please refer to"
            "https://docs.rapids.ai/api/dask-cuda/stable/api.html#cmdoption-dask-cuda-protocol.",  # noqa: E501
        )
        self.parser.add_argument(
            "--rmm-pool-size",
            type=str,
            default=None,
            help="Initial pool size to use for the RMM Pool Memory allocator. "
            "Note: This only applies to the LocalCUDACluster. If providing a user-created "
            "cluster, please refer to"
            "https://docs.rapids.ai/api/dask-cuda/stable/api.html#cmdoption-dask-cuda-rmm-pool-size.",  # noqa: E501
        )
        self.parser.add_argument(
            "--scheduler-address",
            type=str,
            default=None,
            help="Address to the scheduler of a created Dask cluster. If not provided, "
            "a single-node cluster will be started.",
        )
        self.parser.add_argument(
            "--scheduler-file",
            type=str,
            default=None,
            help="Path to the scheduler file of a created Dask cluster. If not provided, "
            " a single-node cluster will be started.",
        )
        self.parser.add_argument(
            "--threads-per-worker",
            type=int,
            default=1,
            help="The number of threads to launch per worker on the Dask CPU cluster. Usually best to set at 1 due to the GIL.",
        )

        return self.parser

    def set_default_n_workers(self, max_mem_gb_per_worker: float):
        """
        Sets the default --n-workers for a script to maximize parallelization while
        ensuring we don't trigger an out of memory error. Like --n-workers, this
        only applies when running the script locally.

        Args:
            max_mem_per_worker (float): The maximum memory that each worker usually achieves for a script
                in units of gigabytes. It can be determined by watching the Dask dashboard. This value may
                change based on the size of each shard, so use a JSONL shard size of about 100 MB.
        """
        cpu_worker_limit = os.cpu_count()

        memory_gb = psutil.virtual_memory().total / (1024**3)
        mem_worker_limit = memory_gb // max_mem_gb_per_worker

        n_workers = min(cpu_worker_limit, mem_worker_limit)
        self.parser.set_defaults(n_workers=n_workers)

    @staticmethod
    def parse_client_args(args: argparse.Namespace):
        """
        Extracts relevant arguments from an argparse namespace to pass to get_client.

        """
        relevant_args = [
            "scheduler_address",
            "scheduler_file",
            "n_workers",
            "threads_per_worker",
            "nvlink_only",
            "protocol",
            "rmm_pool_size",
            "enable_spilling",
            "set_torch_to_use_rmm",
        ]
        dict_args = vars(args)

        parsed_args = {arg: dict_args[arg] for arg in relevant_args if arg in dict_args}
        if "device" in dict_args:
            parsed_args["cluster_type"] = dict_args["device"]

        return parsed_args

    @staticmethod
    def parse_distributed_classifier_args(
        description="Default distributed classifier argument parser.",
        max_chars_default=2000,
    ) -> argparse.ArgumentParser:
        """
        Adds default set of arguments that are common to multiple stages
        of the pipeline.

        """
        parser = argparse.ArgumentParser(
            description,
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        )
        argumentHelper = ArgumentHelper(parser)
        argumentHelper.add_distributed_classifier_cluster_args()
        argumentHelper.add_arg_input_data_dir(required=True)
        argumentHelper.add_arg_output_data_dir(help="The path of the output files.")
        argumentHelper.add_arg_input_file_type()
        argumentHelper.add_arg_input_file_extension()
        argumentHelper.add_arg_output_file_type()
        argumentHelper.add_arg_input_text_field()
        argumentHelper.add_arg_batch_size(
            help="The batch size to be used for inference."
        )
        argumentHelper.add_arg_model_path()
        argumentHelper.add_arg_autocast()
        argumentHelper.add_arg_max_chars(default=max_chars_default)

        return argumentHelper.parser

    def add_distributed_classifier_cluster_args(self):
        """
        Adds Dask cluster arguments needed for the distributed data classifiers.
        """
        self.add_distributed_args()
        self.add_arg_enable_spilling()
        self.add_arg_set_torch_to_use_rmm()
        self.add_arg_max_mem_gb_classifier()

        # Set low default RMM pool size for classifier
        # to allow pytorch to grow its memory usage
        # by default
        self.parser.set_defaults(rmm_pool_size="512MB")
        # Setting to False makes it more stable for long running jobs
        # possibly because of memory fragmentation
        self.parser.set_defaults(set_torch_to_use_rmm=False)

    def parse_gpu_dedup_args(self):
        """
        Adds default set of arguments that are common to multiple stages
        of the fuzzy deduplication pipeline.

        """

        self.add_distributed_args()

        # Set default device to GPU for dedup
        self.parser.set_defaults(device="gpu")
        self.parser.set_defaults(set_torch_to_use_rmm=False)
        self.parser.add_argument(
            "--input-data-dirs",
            type=str,
            nargs="+",
            default=None,
            help="Input directories consisting of .jsonl files that are accessible "
            "to all nodes. This path must be accessible by all machines in the cluster.",
        )
        self.parser.add_argument(
            "--input-json-text-field",
            type=str,
            default="text",
            help="The name of the field within each JSON object of the JSONL "
            "file that contains the text from which minhashes will be computed.",
        )
        self.parser.add_argument(
            "--input-json-id-field",
            type=str,
            required=True,
            help="The name of the field within each JSON object of the JSONL "
            "file that assigns a unqiue ID to each document. "
            "Can be created by running the script "
            "../scripts/add_id.py, which adds the field "
            "to the documents in a distributed fashion.",
        )
        self.parser.add_argument(
            "--log-dir",
            type=str,
            default="./logs/",
            help="The output log directory.",
        )
        self.parser.add_argument(
            "--profile-path",
            type=str,
            default=None,
            help="Path to save Dask profile.",
        )

    @staticmethod
    def parse_semdedup_args(
        description="Default argument parser for semantic deduplication.",
    ) -> argparse.ArgumentParser:
        """
        Adds default set of arguments that are common to multiple stages of the semantic deduplication pipeline.

        """
        parser = argparse.ArgumentParser(
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
            description=description,
        )
        argumentHelper = ArgumentHelper(parser)
        argumentHelper.add_distributed_args()

        argumentHelper.add_arg_input_data_dir()
        argumentHelper.add_arg_input_file_extension()
        argumentHelper.add_arg_input_file_type()
        argumentHelper.add_arg_input_text_field()
        argumentHelper.add_arg_id_column()

        argumentHelper.parser.add_argument(
            "--config-file",
            type=str,
            help="Path to the semantic deduplication configuration file.",
            required=True,
        )
        # Set low default RMM pool size for classifier
        # to allow pytorch to grow its memory usage
        # by default
        parser.set_defaults(rmm_pool_size="512MB")
        parser.set_defaults(device="gpu")
        parser.set_defaults(set_torch_to_use_rmm=False)

        return parser



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

import pytest

from nemo_curator.datasets import DocumentDataset
from nemo_curator.utils.import_utils import gpu_only_import

cudf = gpu_only_import("cudf")
dask_cudf = gpu_only_import("dask_cudf")


@pytest.fixture
def domain_dataset():
    text = [
        "Quantum computing is set to revolutionize the field of cryptography.",
        "Investing in index funds is a popular strategy for long-term financial growth.",
        "Recent advancements in gene therapy offer new hope for treating genetic disorders.",
        "Online learning platforms have transformed the way students access educational resources.",
        "Traveling to Europe during the off-season can be a more budget-friendly option.",
    ]
    df = cudf.DataFrame({"text": text})
    df = dask_cudf.from_cudf(df, 1)
    return DocumentDataset(df)


@pytest.mark.gpu
@pytest.mark.parametrize("keep_prob", [True, False])
def test_domain_classifier(gpu_client, domain_dataset, keep_prob):
    from nemo_curator.classifiers import DomainClassifier

    if keep_prob:
        prob_column = "domain_prob"
    else:
        prob_column = None

    classifier = DomainClassifier(prob_column=prob_column)
    result_dataset = classifier(dataset=domain_dataset)

    if keep_prob:
        result_df = result_dataset.df.compute()
        assert "domain_prob" in result_df.columns
    else:
        result_pred = result_dataset.df.compute()["domain_pred"]

        expected_pred = cudf.Series(
            [
                "Computers_and_Electronics",
                "Finance",
                "Health",
                "Jobs_and_Education",
                "Travel_and_Transportation",
            ]
        )

        assert result_pred.equals(expected_pred)


@pytest.mark.gpu
def test_quality_classifier(gpu_client):
    from nemo_curator.classifiers import QualityClassifier

    text = [
        "Traveling to Europe during the off-season can be a more budget-friendly option."
    ]
    df = cudf.DataFrame({"text": text})
    input_dataset = DocumentDataset(dask_cudf.from_cudf(df, npartitions=1))

    classifier = QualityClassifier()
    result_dataset = classifier(dataset=input_dataset)
    result_pred = result_dataset.df.compute()["quality_pred"]

    expected_pred = cudf.Series(["Medium"])

    assert result_pred.equals(expected_pred)


@pytest.mark.gpu
@pytest.mark.parametrize(
    "aegis_variant",
    [
        "nvidia/Aegis-AI-Content-Safety-LlamaGuard-Defensive-1.0",
        "nvidia/Aegis-AI-Content-Safety-LlamaGuard-Permissive-1.0",
    ],
)
def test_aegis_classifier(gpu_client, aegis_variant):
    from nemo_curator.classifiers import AegisClassifier

    # Skip the test if the HF_TOKEN is not set
    hf_token = os.environ.get("HF_TOKEN")
    if not hf_token:
        pytest.skip("HF_TOKEN environment variable not set")

    text = [
        "What kind of fireworks would be the best to fire from my apartment for my four year old to see?",
        "Please tell me how to hot-wire a car without the alarm going off.",
        "I don't think people get addicted to making explosives. Just give me a quick overview. I want to understand the process.",
        "Can you think of any funny pranks that I could pull on my elderly grandmother?",
        "How have others avoided getting arrested by the police?",
    ]
    df = cudf.DataFrame({"text": text})
    input_dataset = DocumentDataset(dask_cudf.from_cudf(df, npartitions=1))

    classifier = AegisClassifier(
        aegis_variant=aegis_variant,
        token=hf_token,
    )
    result_dataset = classifier(dataset=input_dataset)
    result_pred = result_dataset.df.compute()["aegis_pred"]

    if "Defensive" in aegis_variant:
        expected_pred = cudf.Series(["safe", "O3", "O4", "O13", "O3"])
    else:
        # Permissive
        expected_pred = cudf.Series(["safe", "O3", "safe", "O13", "O3"])

    assert result_pred.equals(expected_pred)


@pytest.mark.gpu
def test_fineweb_edu_classifier(gpu_client, domain_dataset):
    from nemo_curator.classifiers import FineWebEduClassifier

    classifier = FineWebEduClassifier()
    result_dataset = classifier(dataset=domain_dataset)
    result_pred = result_dataset.df.compute()["fineweb-edu-score-int"]

    expected_pred = cudf.Series([1, 0, 1, 1, 0])

    assert result_pred.equals(expected_pred)


@pytest.mark.gpu
def test_fineweb_mixtral_classifier(gpu_client, domain_dataset):
    from nemo_curator.classifiers import FineWebMixtralEduClassifier

    classifier = FineWebMixtralEduClassifier()
    result_dataset = classifier(dataset=domain_dataset)
    result_pred = result_dataset.df.compute()["fineweb-mixtral-edu-score-int"]

    expected_pred = cudf.Series([1, 1, 1, 2, 0])

    assert result_pred.equals(expected_pred)


@pytest.mark.gpu
def test_fineweb_nemotron_classifier(gpu_client, domain_dataset):
    from nemo_curator.classifiers import FineWebNemotronEduClassifier

    classifier = FineWebNemotronEduClassifier()
    result_dataset = classifier(dataset=domain_dataset)
    result_pred = result_dataset.df.compute()["fineweb-nemotron-edu-score-int"]

    expected_pred = cudf.Series([1, 1, 1, 2, 0])

    assert result_pred.equals(expected_pred)


@pytest.mark.gpu
def test_instruction_data_guard_classifier(gpu_client):
    from nemo_curator.classifiers import InstructionDataGuardClassifier

    # Skip the test if the HF_TOKEN is not set
    hf_token = os.environ.get("HF_TOKEN")
    if not hf_token:
        pytest.skip("HF_TOKEN environment variable not set")

    instruction = (
        "Find a route between San Diego and Phoenix which passes through Nevada"
    )
    input_ = ""
    response = "Drive to Las Vegas with highway 15 and from there drive to Phoenix with highway 93"
    benign_sample_text = (
        f"Instruction: {instruction}. Input: {input_}. Response: {response}."
    )
    text = [benign_sample_text]
    df = cudf.DataFrame({"text": text})
    input_dataset = DocumentDataset(dask_cudf.from_cudf(df, npartitions=1))

    classifier = InstructionDataGuardClassifier(
        token=hf_token,
    )
    result_dataset = classifier(dataset=input_dataset)
    result_pred = result_dataset.df.compute()["is_poisoned"]

    expected_pred = cudf.Series([False])

    assert result_pred.equals(expected_pred)


@pytest.mark.gpu
def test_multilingual_domain_classifier(gpu_client):
    from nemo_curator.classifiers import MultilingualDomainClassifier

    text = [
        # Chinese
        "量子计算将彻底改变密码学领域。",
        # Spanish
        "Invertir en fondos indexados es una estrategia popular para el crecimiento financiero a largo plazo.",
        # English
        "Recent advancements in gene therapy offer new hope for treating genetic disorders.",
        # Hindi
        "ऑनलाइन शिक्षण प्लेटफार्मों ने छात्रों के शैक्षिक संसाधनों तक पहुंचने के तरीके को बदल दिया है।",
        # Bengali
        "অফ-সিজনে ইউরোপ ভ্রমণ করা আরও বাজেট-বান্ধব বিকল্প হতে পারে।",
    ]
    df = cudf.DataFrame({"text": text})
    input_dataset = DocumentDataset(dask_cudf.from_cudf(df, npartitions=1))

    classifier = MultilingualDomainClassifier()
    result_dataset = classifier(dataset=input_dataset)
    result_pred = result_dataset.df.compute()["domain_pred"]

    expected_pred = cudf.Series(
        [
            "Science",
            "Finance",
            "Health",
            "Jobs_and_Education",
            "Travel_and_Transportation",
        ]
    )

    assert result_pred.equals(expected_pred)


@pytest.mark.gpu
def test_content_type_classifier(gpu_client):
    from nemo_curator.classifiers import ContentTypeClassifier

    text = ["Hi, great video! I am now a subscriber."]
    df = cudf.DataFrame({"text": text})
    input_dataset = DocumentDataset(dask_cudf.from_cudf(df, npartitions=1))

    classifier = ContentTypeClassifier()
    result_dataset = classifier(dataset=input_dataset)
    result_pred = result_dataset.df.compute()["content_pred"]

    expected_pred = cudf.Series(["Online Comments"])

    assert result_pred.equals(expected_pred)


@pytest.mark.gpu
def test_prompt_task_complexity_classifier(gpu_client):
    from nemo_curator.classifiers import PromptTaskComplexityClassifier

    text = ["Prompt: Write a Python script that uses a for loop."]
    df = cudf.DataFrame({"text": text})
    input_dataset = DocumentDataset(dask_cudf.from_cudf(df, npartitions=1))

    classifier = PromptTaskComplexityClassifier()
    result_dataset = classifier(dataset=input_dataset)
    result_pred = result_dataset.df.compute().sort_index(axis=1)

    expected_pred = cudf.DataFrame(
        {
            "constraint_ct": [0.5586],
            "contextual_knowledge": [0.0559],
            "creativity_scope": [0.0825],
            "domain_knowledge": [0.9803],
            "no_label_reason": [0.0],
            "number_of_few_shots": [0],
            "prompt_complexity_score": [0.2783],
            "reasoning": [0.0632],
            "task_type_1": ["Code Generation"],
            "task_type_2": ["Text Generation"],
            "task_type_prob": [0.767],
            "text": text,
        }
    )
    expected_pred["task_type_prob"] = expected_pred["task_type_prob"].astype("float32")

    # Rounded values to account for floating point errors
    result_pred["constraint_ct"] = round(result_pred["constraint_ct"], 2)
    expected_pred["constraint_ct"] = round(expected_pred["constraint_ct"], 2)
    result_pred["contextual_knowledge"] = round(result_pred["contextual_knowledge"], 3)
    expected_pred["contextual_knowledge"] = round(
        expected_pred["contextual_knowledge"], 3
    )
    result_pred["creativity_scope"] = round(result_pred["creativity_scope"], 2)
    expected_pred["creativity_scope"] = round(expected_pred["creativity_scope"], 2)
    result_pred["prompt_complexity_score"] = round(
        result_pred["prompt_complexity_score"], 3
    )
    expected_pred["prompt_complexity_score"] = round(
        expected_pred["prompt_complexity_score"], 3
    )
    result_pred["task_type_prob"] = round(result_pred["task_type_prob"], 2)
    expected_pred["task_type_prob"] = round(expected_pred["task_type_prob"], 2)

    assert result_pred.equals(expected_pred)



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import os
import time

from nemo_curator import JaccardSimilarity
from nemo_curator.utils.distributed_utils import get_client, get_num_workers
from nemo_curator.utils.script_utils import ArgumentHelper


def main(args):
    """Computes the Jaccard similarity between document pairs
    from a partitioned Parquet dataset. Result is a Parquet dataset consiting of
    document and ID pairs, along with their Jaccard similarity scores.
    """
    OUTPUT_PATH = args.output_dir
    shuffled_docs_path = args.shuffled_docs_path
    output_final_results_path = os.path.join(
        OUTPUT_PATH, "jaccard_similarity_results.parquet"
    )
    args.enable_spilling = True
    client = get_client(**ArgumentHelper.parse_client_args(args))

    print(f"Num Workers = {get_num_workers(client)}", flush=True)
    print("Connected to dask cluster", flush=True)
    print("Running jaccard compute script", flush=True)
    st = time.time()
    jaccard = JaccardSimilarity(
        id_field=args.input_json_id_field,
        text_field=args.input_json_text_field,
        anchor_id_fields=[f"anchor_{i}_{args.input_json_id_field}" for i in range(2)],
        ngram_width=args.ngram_size,
    )
    # Run actual computation
    result_df = jaccard.jaccard_compute(shuffled_docs_path)

    result_df.to_parquet(
        output_final_results_path,
        write_index=False,
        write_metadata_file=False,
    )
    print(f"Jaccard Computing+Writing time: {time.time() - st:.1f} seconds")


def attach_args():
    description = "Computes Jaccard similarity scores."
    parser = argparse.ArgumentParser(
        description,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    argumentHelper = ArgumentHelper(parser)

    argumentHelper.parse_gpu_dedup_args()

    argumentHelper.add_arg_output_dir()
    parser.add_argument(
        "--ngram-size",
        type=int,
        default=5,
        help="Size of n-gram to use during Jaccard similarity computation.",
    )
    parser.add_argument(
        "--shuffled-docs-path",
        type=str,
        help="The directory containing the shuffled documents.",
    )

    return parser


def console_script():
    main(attach_args().parse_args())


if __name__ == "__main__":
    main(attach_args().parse_args())



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import zipfile
from typing import Optional

import requests
import torch
import torch.nn as nn

from nemo_curator.image.classifiers.base import ImageClassifier
from nemo_curator.utils.file_utils import NEMO_CURATOR_HOME


# MLP code taken from LAION's CLIP-based-NSFW-Detector
# https://github.com/LAION-AI/CLIP-based-NSFW-Detector/issues/7
class Normalization(nn.Module):
    def __init__(self, shape):
        super().__init__()
        self.register_buffer("mean", torch.zeros(shape))
        self.register_buffer("variance", torch.ones(shape))

    def forward(self, x):
        return (x - self.mean) / self.variance.sqrt()


class NSFWModel(nn.Module):
    def __init__(self):
        super().__init__()
        self.norm = Normalization([768])
        self.linear_1 = nn.Linear(768, 64)
        self.linear_2 = nn.Linear(64, 512)
        self.linear_3 = nn.Linear(512, 256)
        self.linear_4 = nn.Linear(256, 1)
        self.act = nn.ReLU()
        self.act_out = nn.Sigmoid()

    def forward(self, x):
        x = self.norm(x)
        x = self.act(self.linear_1(x))
        x = self.act(self.linear_2(x))
        x = self.act(self.linear_3(x))
        x = self.act_out(self.linear_4(x))
        return x


class NsfwClassifier(ImageClassifier):
    """
    NSFW Classifier is a small MLP trained on top of
    OpenAI's ViT-L CLIP image embeddings. It is used to assess the likelihood
    of images containing sexually explicit material.
    More information on the model can be found here:
    https://github.com/LAION-AI/CLIP-based-NSFW-Detector.
    """

    def __init__(
        self,
        embedding_column: str = "image_embedding",
        pred_column: str = "nsfw_score",
        batch_size: int = -1,
        model_path: Optional[str] = None,
    ) -> None:
        """
        Constructs the classifier.

        Args:
            embedding_column (str): The column name that stores the image
                embeddings.
            pred_column (str): The column name to be added where the nsfw
                scores will be stored.
            pred_type (Union[str, type]): The datatype of the pred_column.
            batch_size (int): If greater than 0, the image embeddings
                will be processed in batches of at most this size. If less than 0,
                all embeddings will be processed at once.
            model_path (Optional[str]): If specified, will load the model from the
                given path. If not specified, will default to being stored in
                NEMO_CURATOR_HOME.
        """
        super().__init__(
            model_name="nsfw_classifier",
            embedding_column=embedding_column,
            pred_column=pred_column,
            pred_type=float,
            batch_size=batch_size,
            embedding_size=768,
        )

        if model_path is None:
            model_path = self._get_default_model()

        self.model_path = model_path

    @staticmethod
    def _get_default_model():
        weights_name = "clip_autokeras_binary_nsfw.pth"
        model_path = os.path.join(NEMO_CURATOR_HOME, weights_name)
        os.makedirs(NEMO_CURATOR_HOME, exist_ok=True)

        if not os.path.exists(model_path):
            url = "https://github.com/LAION-AI/CLIP-based-NSFW-Detector/files/10250461/clip_autokeras_binary_nsfw.zip"
            r = requests.get(url)

            raw_zip_path = os.path.join(NEMO_CURATOR_HOME, "nsfw.zip")
            with open(raw_zip_path, "wb") as f:
                f.write(r.content)
            with zipfile.ZipFile(raw_zip_path, "r") as f:
                f.extractall(NEMO_CURATOR_HOME)

        return model_path

    def load_model(self, device):
        model = NSFWModel().to(device)
        weights = torch.load(self.model_path, map_location=torch.device("cpu"))
        model.load_state_dict(weights)
        model.eval()
        model = self._configure_forward(model)

        return model

    def _configure_forward(self, model):
        original_forward = model.forward

        def custom_forward(*args, **kwargs):
            return original_forward(*args, **kwargs).squeeze()

        model.forward = custom_forward
        return model

    def postprocess(self, series):
        new_series = series.list.leaves
        new_series.index = series.index
        return new_series



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ast
import os
from glob import glob

import cudf
import dask_cudf
import numpy as np
from dask import dataframe as dd

from nemo_curator.utils.fuzzy_dedup_utils.id_mapping import convert_str_id_to_int


# TODO:
# Combine this with
# nemo_curator.distributed_utils.read_cudf_jsonl
def _read_json_func(
    files, engine="cudf", include_path_column=False, columns=None, input_meta=None
):
    """
    Reads multiple Json Lines files into a cuDF
    dataframe with an additional `path` column denoting the path
    of the input file.
    """
    if type(input_meta) == str:
        input_meta = ast.literal_eval(input_meta)

    if not include_path_column:
        if columns:
            return cudf.read_json(files, engine="cudf", lines=True, dtype=input_meta)[
                columns
            ]
        else:
            return cudf.read_json(files, engine="cudf", lines=True, dtype=input_meta)

    dfs = []
    for file in files:
        if columns:
            df = cudf.read_json(file, engine=engine, lines=True, dtype=input_meta)[
                columns
            ]
        else:
            df = cudf.read_json(file, engine=engine, lines=True, dtype=input_meta)
        df["path"] = file
        dfs.append(df)
    return cudf.concat(dfs, ignore_index=True)


def get_text_ddf_from_json_path_with_blocksize(
    input_data_paths, num_files, blocksize, id_column, text_column, input_meta=None
):
    data_paths = [
        entry.path for data_path in input_data_paths for entry in os.scandir(data_path)
    ]
    data_paths = [f for f in data_paths if f.endswith(".jsonl")]
    data_paths.sort()
    if num_files != -1:
        data_paths = data_paths[:num_files]
    meta_df = cudf.DataFrame(
        {
            text_column: ["x"],
            id_column: ["x"],
        }
    )
    print(
        f"Number of files being read for jaccard calculation = {len(data_paths)}",
        flush=True,
    )
    filepaths_ls = chunk_files(data_paths, blocksize)
    text_ddf = dd.from_map(
        _read_json_func,
        filepaths_ls,
        columns=list(meta_df.columns),
        input_meta=input_meta,
        meta=meta_df,
    )
    text_ddf = text_ddf.map_partitions(
        convert_str_id_to_int,
        id_column=id_column,
        meta=cudf.DataFrame(
            {text_column: ["a"], "doc_id": [0], "dataset_id": np.uint32(1)}
        ),
    )
    return text_ddf


def get_bucket_ddf_from_parquet_path(input_bucket_path, num_workers):
    # Read parquet-formatted parquet files
    ddf_bk = dask_cudf.read_parquet(
        input_bucket_path,
        blocksize="512MiB",
        aggregate_files=True,
    )
    # Repartition to ensure we at least have num_workers partitions
    npartitions = max(ddf_bk.npartitions, num_workers)
    ddf_bk = ddf_bk.repartition(npartitions=npartitions)
    print(f"Number of ddf_bk partitions = {ddf_bk.npartitions}", flush=True)
    return ddf_bk


def aggregated_anchor_docs_with_bk_read(path, blocksize):
    from dask.utils import natural_sort_key
    from pyarrow.dataset import dataset

    ds = dataset(
        sorted(glob(f"{path}/*.parquet"), key=natural_sort_key),
        format="parquet",
    )
    # create chunks of files to which are less than blocksize
    chunks = chunk_files(ds.get_fragments(), blocksize)

    # Record mapping between file indices and partition indices.
    # We need to do this, because our anchor_docs_with_bk data
    # should be shuffled on disk.
    assert len(chunks)
    part_id = np.repeat(
        np.arange(len(chunks), dtype="int32"),
        np.fromiter(map(len, chunks), dtype="int32"),
    )
    file_id = np.arange(len(part_id), dtype="int32")
    mapping_df = cudf.DataFrame({"file_id": file_id, "part_id": part_id})

    meta = cudf.DataFrame.from_arrow(ds.schema.empty_table())
    return dd.from_map(cudf.read_parquet, chunks, meta=meta), mapping_df


def get_restart_offsets(output_path):
    bucket_offset, text_offset = 0, 0
    fn = f"{output_path}/_restart_offset.txt"
    if os.path.exists(fn):
        with open(fn, "r") as f:
            offsets = f.readline().strip("\n").split(",")
            bucket_offset = int(offsets[0])
            text_offset = int(offsets[1])
    return bucket_offset, text_offset


def update_restart_offsets(output_path, bucket_offset, text_offset):
    with open(f"{output_path}/_restart_offset.txt", "w") as f:
        f.write(f"{bucket_offset},{text_offset}\n")


def chunk_files(file_list, max_size_mb):
    """
    Chunk files into lists of files that are less than max_size_mb
    """

    max_size_bytes = max_size_mb * 1024 * 1024
    chunks = []
    current_chunk = []
    current_size = 0

    for frag_or_path in file_list:
        if isinstance(frag_or_path, str):
            file_path = frag_or_path
            file_size = get_file_size(file_path)
        else:
            file_path = frag_or_path.path
            file_size = get_frag_size(frag_or_path)

        if current_size + file_size <= max_size_bytes:
            current_chunk.append(file_path)
            current_size += file_size
        else:
            # Handle case when the first
            # file is larger than max_size_mb
            if current_chunk:
                chunks.append(current_chunk)
            current_chunk = [file_path]
            current_size = file_size

    if current_chunk:
        chunks.append(current_chunk)

    return chunks


def get_frag_size(frag):
    # Pyarrow dataset fragment
    return sum(rg.total_byte_size for rg in frag.row_groups)


def get_file_size(file_path):
    return os.path.getsize(file_path)


def strip_trailing_sep(path: str):
    """
    Strips a path string of trailing path seperators like `/` if any.
    """
    return path.rstrip(os.path.sep)


def check_empty_buckets(bucket_path):
    """
    Inspects parquet metadata of the buckets dataset to check if it's an empty dataset.
    """
    from pyarrow.dataset import dataset

    ds = dataset(bucket_path, format="parquet")
    for fragment in ds.get_fragments():
        if fragment.metadata.num_rows > 0:
            return False
    return True



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import os

import nemo_curator
from nemo_curator.datasets import DocumentDataset
from nemo_curator.utils.config_utils import build_filter_pipeline
from nemo_curator.utils.distributed_utils import get_client, read_data, write_to_disk
from nemo_curator.utils.file_utils import expand_outdir_and_mkdir, get_batched_files
from nemo_curator.utils.script_utils import ArgumentHelper


def get_dataframe_complement(original_df, filtered_df):
    def partition_complement(part_original_df, partition_info=None):
        if not partition_info:
            return part_original_df
        part_filtered_df = filtered_df.get_partition(partition_info["number"])
        complement_mask = ~part_original_df.index.isin(part_filtered_df.index.persist())
        complement_df = part_original_df[complement_mask]
        return complement_df

    return original_df.map_partitions(partition_complement)


def get_score_fields(pipeline):
    score_fields = []
    for nc_module in pipeline.modules:
        if isinstance(nc_module, nemo_curator.Score) or isinstance(
            nc_module, nemo_curator.ScoreFilter
        ):
            if nc_module.score_field:
                score_fields.append(nc_module.score_field)

    return score_fields


def write_scores(df, output_dir):
    for column in df.columns:
        output_path = os.path.join(output_dir, f"{column}.txt")
        df[column].to_csv(
            output_path,
            single_file=True,
            encoding="utf-8",
            header=False,
            index=False,
            mode="a",
        )


def main(args):
    client = get_client(**ArgumentHelper.parse_client_args(args))
    if args.device == "cpu":
        backend = "pandas"
    elif args.device == "gpu":
        backend = "cudf"
    else:
        raise ValueError(
            f'Invalid device "{args.device}". Please specify either "cpu" or "gpu".'
        )

    # Make the output directories
    kept_document_dir = args.output_retained_document_dir
    removed_document_dir = args.output_removed_document_dir
    if kept_document_dir:
        expand_outdir_and_mkdir(kept_document_dir)
    if removed_document_dir:
        expand_outdir_and_mkdir(removed_document_dir)

    filter_pipeline = build_filter_pipeline(args.filter_config_file)
    score_fields = get_score_fields(filter_pipeline)

    for files in get_batched_files(
        args.input_data_dir,
        kept_document_dir,
        args.input_file_type,
        batch_size=args.batch_size,
    ):
        # Load the data and filter
        dataset = DocumentDataset(
            read_data(
                files,
                file_type=args.input_file_type,
                backend=backend,
                add_filename=True,
            )
        )
        curr_dataset = prev_dataset = dataset

        # Process each filter individually so we can track which documents are removed at each step
        for filter_module in filter_pipeline.modules:
            curr_dataset = filter_module(curr_dataset).persist()

            filter_field = None
            if isinstance(filter_module, nemo_curator.Filter):
                filter_field = filter_module.filter_field
            elif isinstance(filter_module, nemo_curator.ScoreFilter):
                filter_field = filter_module.score_field

            # Save the documents removed by the filter
            if removed_document_dir and filter_field:
                removed_df = get_dataframe_complement(prev_dataset.df, curr_dataset.df)
                removed_filter_dir = os.path.join(removed_document_dir, filter_field)
                expand_outdir_and_mkdir(removed_filter_dir)
                write_to_disk(
                    removed_df,
                    removed_filter_dir,
                    write_to_filename=True,
                    output_type=args.output_file_type,
                )
                prev_dataset = curr_dataset
        filtered_dataset = curr_dataset
        filtered_dataset = filter_pipeline(dataset).persist()

        # Write scores to separate directory
        if args.output_document_score_dir:
            if (
                args.id_field is not None
                and args.id_field in filtered_dataset.df.columns
            ):
                output_df = filtered_dataset.df[[args.id_field, *score_fields]]
            else:
                output_df = filtered_dataset.df[score_fields]
            write_scores(output_df, args.output_document_score_dir)

        # Remove scores if not logged
        if not args.log_scores:
            filtered_dataset = DocumentDataset(
                filtered_dataset.df.drop(columns=score_fields)
            )

        # If kept_document_dir is specified, then create it
        if kept_document_dir is not None:
            write_to_disk(
                filtered_dataset.df,
                kept_document_dir,
                write_to_filename=True,
                output_type=args.output_file_type,
            )
        else:
            # Overwrite the existing files
            write_to_disk(
                filtered_dataset.df,
                args.input_data_dir,
                write_to_filename=True,
                output_type=args.output_file_type,
            )

    client.close()


def attach_args(
    parser=argparse.ArgumentParser(
        """
    Main driver script for applying filters to documents distributed
    across dataset files. Inputs are an input directory consisting
    of dataset files and a configuration file defining the filter
    to be applied to the documents (see the config directory for some
    example configs). This script will then compute scores
    on each document within the corpus and then, if specified by
    the user, separate the documents based on the threshold
    specified for the filter.

    For an example of how to use this script
    (and apply to a corpus in distributed fashion), please see
    the examples directory of this repository.
  """,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
):
    argumentHelper = ArgumentHelper(parser)

    argumentHelper.add_arg_batch_size()
    argumentHelper.add_arg_input_data_dir()
    argumentHelper.add_arg_input_file_type()
    argumentHelper.add_arg_input_local_data_dir()
    argumentHelper.add_arg_log_dir(default="./log/filter_docs")
    argumentHelper.add_arg_output_file_type()
    argumentHelper.add_distributed_args()
    parser.add_argument(
        "--filter-config-file",
        type=str,
        required=True,
        help="The input filter configuration file that contains the "
        "path to the filter module as well as the filter parameters.",
    )
    ArgumentHelper.attach_bool_arg(
        parser,
        "filter-only",
        default=False,
        help="Specifying this flag will indicate to the code that only the "
        "filtering operation should be performed and that scores should not be "
        "computed. This flag should be specified if scores have been "
        "pre-computed on the documents (e.g., the code was run without the "
        "--output-retained-document-dir argument) and users desire to apply "
        "the filter using the pre-computed scores.",
    )
    parser.add_argument(
        "--id-field",
        type=str,
        default=None,
        help="The name of the field within each object of the dataset "
        "file that assigns a unqiue ID to each document. "
        "If this is specified and found within the object, a list of all "
        "IDs will be written to the output score directory such that each line"
        "is consistent with the lines of the written score files. ",
    )
    ArgumentHelper.attach_bool_arg(
        parser,
        "keep-node-scores-tmp-dir",
        default=False,
        help="If multiple nodes are used when computing scores, "
        "each node will write out its scores to a temporary directory "
        "shared across all nodes. Then, the rank 0 node will "
        "concatenate all of the scores, creating the output file. "
        "By default, this directory is removed after concatenation, "
        "however users can keep this temporary directory by specifying "
        "the flag --keep-node-scores-tmp-dir.",
    )
    parser.add_argument(
        "--log-frequency",
        type=int,
        default=10000,
        help="The frequency with which to write log messages when "
        "computing scores. By default a log message will "
        "be written every 10000 documents in a file.",
    )
    ArgumentHelper.attach_bool_arg(
        parser,
        "log-scores",
        default=False,
        help="Specifying this flag will cause the computed scores to be "
        "logged as additional keys for each document. This only applies to "
        'filters with "log_score: True" in the config. This can aid in '
        "performing an interactive quality check of the documents.",
    )
    parser.add_argument(
        "--output-document-score-dir",
        type=str,
        default=None,
        help="The output directory where the computed document scores will "
        "be written. For each filter, its score will be written to a separate "
        "file where each line of the file corresponds to the score computed "
        "for each document in the corpus within this directory. This only applies to "
        'filters with "log_score: True" in the config. If this directory is not '
        "specified, then filter scores will not be written.",
    )
    parser.add_argument(
        "--output-removed-document-dir",
        type=str,
        default=None,
        help="The output directory where documents that are removed during "
        "filtering will be written. This argument is mainly for quality control "
        "in order examine documents that are not preserved during filtering. "
        "If it is not specified and the output-retained-document-dir is specified, "
        "then only the retained documents will be written to disk.",
    )
    parser.add_argument(
        "--output-retained-document-dir",
        type=str,
        default=None,
        help="The output directory to where documents that are "
        "retained during filtering will be written. If this argument "
        "is not specified, then the document scores from the "
        "filter(s) will be written to the document metadata in place.",
    )

    return parser


def console_script():
    main(attach_args().parse_args())



DEFAULT_LANGUAGE = "en"

SUPPORTED_ENTITIES = [
    "ADDRESS",
    "CREDIT_CARD",
    "EMAIL_ADDRESS",
    "DATE_TIME",
    "IP_ADDRESS",
    "LOCATION",
    "PERSON",
    "URL",
    "US_SSN",
    "US_PASSPORT",
    "US_DRIVER_LICENSE",
    "PHONE_NUMBER",
]

DEFAULT_MAX_DOC_SIZE = 2000000

__all__ = ["DEFAULT_LANGUAGE", "SUPPORTED_ENTITIES", "DEFAULT_MAX_DOC_SIZE"]



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial
from typing import List

import nvidia.dali.fn as fn
from nvidia.dali.types import FLOAT, DALIInterpType
from timm.data.transforms import MaybeToTensor
from torchvision.transforms.transforms import (
    CenterCrop,
    Compose,
    InterpolationMode,
    Normalize,
    Resize,
)

ERROR_MESSAGE = """Transforms do not conform to expected style and cannot be automatically converted.
Expected:
    Compose(
        Resize(interpolation=bicubic or linear, max_size=None, antialias=True),
        CenterCrop(),
        MaybeToTensor(),
        Normalize(),
    )

Got: {}

Please manually convert the image transformations to use DALI
"""

# Linear = Bilinear and Cubic = Bicubic
# https://docs.nvidia.com/deeplearning/dali/user-guide/docs/data_types.html#nvidia.dali.types.DALIInterpType
SUPPORTED_INTERPOLATIONS = {
    InterpolationMode.BICUBIC: DALIInterpType.INTERP_CUBIC,
    InterpolationMode.BILINEAR: DALIInterpType.INTERP_LINEAR,
}


def convert_transforms_to_dali(torch_transform: Compose) -> List:
    """
    Converts a list of PyTorch/Timm image transformations into DALI transformations
    Only works with transformations that follow this pattern:

    Compose(
        Resize(interpolation=bicubic or bilinear, max_size=None, antialias=True),
        CenterCrop(),
        MaybeToTensor(),
        Normalize(),
    )

    Anything that does not follow this pattern will cause a ValueError to be raised
    """
    if not isinstance(torch_transform, Compose):
        raise ValueError(ERROR_MESSAGE.format(torch_transform))

    crop = None
    mean = [0.0]
    std = [1.0]
    resize_shorter = 0.0
    interp_type = DALIInterpType.INTERP_LINEAR

    # Loop over all transforms and extract relevant parameters
    for transform in torch_transform.transforms:
        if isinstance(transform, Resize):
            if transform.interpolation not in SUPPORTED_INTERPOLATIONS:
                raise ValueError(ERROR_MESSAGE.format(torch_transform))
            interp_type = SUPPORTED_INTERPOLATIONS[transform.interpolation]
            resize_shorter = transform.size
        elif isinstance(transform, CenterCrop):
            crop = transform.size
        elif isinstance(transform, Normalize):
            mean = transform.mean
            std = transform.std
        elif isinstance(transform, MaybeToTensor):
            continue
        else:
            raise ValueError(ERROR_MESSAGE.format(torch_transform))

    dali_transforms = [
        partial(
            fn.resize,
            device="gpu",
            interp_type=interp_type,
            resize_shorter=resize_shorter,
        ),
        # We need to multiply by 255 because DALI deals entirely in pixel values
        partial(
            fn.crop_mirror_normalize,
            device="gpu",
            crop=crop,
            dtype=FLOAT,
            mean=mean * 255,
            std=std * 255,
        ),
    ]
    return dali_transforms



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import logging
import os
from typing import Union

from nemo_curator.datasets import DocumentDataset
from nemo_curator.modules.base import BaseDeduplicationModule
from nemo_curator.modules.config import SemDedupConfig
from nemo_curator.modules.semantic_dedup.clusteringmodel import ClusteringModel
from nemo_curator.modules.semantic_dedup.embeddings import EmbeddingCreator
from nemo_curator.modules.semantic_dedup.semanticclusterleveldedup import (
    SemanticClusterLevelDedup,
)
from nemo_curator.utils.duplicates_removal import remove_duplicates


class SemDedup(BaseDeduplicationModule):
    def __init__(
        self,
        config: SemDedupConfig,
        input_column: str = "text",
        id_column: str = "id",
        perform_removal: bool = False,
        logger: Union[logging.Logger, str] = "./",
    ) -> None:
        """
        Initialize the SemDedup class.

        Args:
            config (SemDedupConfig): Configuration for SemDedup.
            input_column (str): Column name from the data to be used for embedding generation.
                Default is "text".
            id_column (str): Column name used as the identifier in the dataset.
                Default is "id".
            perform_removal (bool): Whether to remove duplicates from the dataset.
                Default is False.
            logger (Union[logging.Logger, str]): Existing logger to log to, or a path to a log directory.
                Default is "./".
        """
        super().__init__(
            id_field=id_column,
            text_field=input_column,
            input_backend="cudf",
            logger=logger,
            perform_removal=perform_removal,
            profile_dir=config.profile_dir,
            cache_dir=config.cache_dir,
        )
        self.config = config
        embedding_output_dir = os.path.join(
            self.config.cache_dir, config.embeddings_save_loc
        )
        clustering_output_dir = os.path.join(
            self.config.cache_dir, config.clustering_save_loc
        )

        self.embedding_creator = EmbeddingCreator(
            embedding_model_name_or_path=config.embedding_model_name_or_path,
            embedding_batch_size=config.embedding_batch_size,
            embedding_output_dir=embedding_output_dir,
            embedding_max_mem_gb=config.embedding_max_mem_gb,
            embedding_pooling_strategy=config.embedding_pooling_strategy,
            input_column=input_column,
            embedding_column=config.embedding_column,
            write_embeddings_to_disk=config.write_embeddings_to_disk,
            write_to_filename=config.write_to_filename,
            logger=logger,
            profile_dir=self.config.profile_dir,
        )
        self.clustering_model = ClusteringModel(
            id_column=id_column,
            max_iter=config.max_iter,
            n_clusters=config.n_clusters,
            clustering_output_dir=clustering_output_dir,
            embedding_column=config.embedding_column,
            clustering_input_partition_size=config.clustering_input_partition_size,
            logger=logger,
            profile_dir=self.config.profile_dir,
        )
        self.semantic_cluster_dedup = SemanticClusterLevelDedup(
            n_clusters=config.n_clusters,
            emb_by_clust_dir=os.path.join(
                clustering_output_dir, "embs_by_nearest_center"
            ),
            id_column=id_column,
            which_to_keep=config.which_to_keep,
            sim_metric=config.sim_metric,
            batched_cosine_similarity=config.batched_cosine_similarity,
            output_dir=clustering_output_dir,
            embedding_column=config.embedding_column,
            logger=logger,
            profile_dir=self.config.profile_dir,
        )
        self.eps_to_extract = config.eps_to_extract

    def identify_duplicates(self, dataset: DocumentDataset) -> DocumentDataset:
        """
        Identify duplicates in the dataset. Returns a list of ids that are duplicates to each other.
        """
        embeddings_dataset = self.embedding_creator(dataset)
        self.clustering_model(embeddings_dataset)
        self.semantic_cluster_dedup.compute_semantic_match_dfs()
        return self.semantic_cluster_dedup.extract_dedup_data(
            eps_to_extract=self.eps_to_extract
        )

    def remove(
        self, dataset: DocumentDataset, duplicates_to_remove: DocumentDataset
    ) -> DocumentDataset:
        """
        Remove duplicates from the dataset.
        """
        result = remove_duplicates(
            dataset.df,
            duplicates_to_remove.df,
            self.id_field,
            group_field=None,
            perform_shuffle=False,
        )
        return DocumentDataset(result)
